from .probability import ProbabilityEstimator
from .feature_importances import FIExtractor
from .dataframe_transformer import DataFrameTransformer
from .country_by_country_regressions import CountryByCountryRegression

__all__ = [
    "DataFrameTransformer",
    "ProbabilityEstimator",
    "FIExtractor",
    "CountryByCountryRegression"
]


import warnings
from typing import List, Optional

import numpy as np
import pandas as pd

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management import reduce_df


def extend_history(
    df: pd.DataFrame,
    new_xcat: str,
    cids: Optional[List[str]] = None,
    hierarchy: List[str] = [],
    backfill: bool = False,
    start: str = None,
):
    """
    Extends the history of a dataframe by creating a new xcat by combining hierarchical categories.
    The method prioritizes superior categories for the new xcat and supplements with inferior ones
    where superior category data is unavailable.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe containing categories that are to be extended.
    new_xcat : str
        The name of the new xcat.
    cids : List[str], optional
        The cross sections to extend. If None, all cids available for any category in 'hierarchy' are extended.
    hierarchy : List[str]
         list of categories from best to worst for representation of the concept.
         Inferior categories are only used to extend the history of the superior ones.
         The new category consists of the best representation category values
         and inferior category values that are available prior to any superior.
    backfill : bool, optional
        If True, the new xcat is backfilled to the start date specified by the 'start' parameter.
    start : str, optional
        The start date of the new xcat. If backfill is True, this values will be backfilled up to this date.

    Returns
    -------
    ~pandas.DataFrame
        standardized DataFrame for the new xcat with extended history, with the columns:
        'cid', 'xcat', 'real_date' and 'value'.
    """

    df = QuantamentalDataFrame(df)
    result_as_categorical = df.InitializedAsCategorical
    _extend_history_checks(
        df=df,
        new_xcat=new_xcat,
        cids=cids,
        hierarchy=hierarchy,
        backfill=backfill,
        start=start,
    )
    start = pd.to_datetime(start)

    extended_results = []

    df, _, cids_in_df = reduce_df(df=df, xcats=hierarchy, cids=cids, out_all=True)

    if df.empty:
        raise ValueError("No data available for the specified cids and categories.")

    if cids is None:
        cids = cids_in_df
    else:
        missing_cids = list(set(cids) - set(cids_in_df))
        cids = cids_in_df
        if len(missing_cids) > 0:
            warnings.warn(
                f"Warning: cids {missing_cids} do not exist for any category in hierarchy. They will be ignored."
            )

    for cid in cids:

        cid_df = df[df["cid"] == cid]

        extended_series = pd.DataFrame()

        for category in hierarchy:

            cat_df = cid_df[cid_df["xcat"] == category].sort_values("real_date")

            if extended_series.empty:
                extended_series = cat_df.copy()
            else:
                min_real_date = extended_series["real_date"].min()

                inferior_values = cat_df[cat_df["real_date"] < min_real_date]
                extended_series = pd.concat([extended_series, inferior_values])

        extended_series = extended_series.sort_values("real_date")

        extended_series["xcat"] = new_xcat
        extended_series["cid"] = cid

        if backfill:
            if not extended_series.empty:
                min_date = extended_series["real_date"].min()
                if min_date > start:
                    backfilled_data = pd.DataFrame(
                        {
                            "real_date": pd.bdate_range(
                                start=start, end=min_date - pd.Timedelta(days=1)
                            ),
                            "value": extended_series.iloc[0]["value"],
                            "cid": cid,
                            "xcat": new_xcat,
                        }
                    )
                    extended_series = pd.concat([backfilled_data, extended_series])
        elif start is not None:
            extended_series = extended_series[extended_series["real_date"] >= start]

        # Add new_xcat and cid
        extended_series["xcat"] = new_xcat
        extended_series["cid"] = cid

        extended_results.append(extended_series)

    extended_df = pd.concat(extended_results, ignore_index=True)
    extended_df = extended_df.sort_values(["cid", "real_date"])

    return QuantamentalDataFrame(extended_df, categorical=result_as_categorical)


def _extend_history_checks(
    df: pd.DataFrame,
    new_xcat: str,
    cids: Optional[List[str]] = None,
    hierarchy: List[str] = [],
    backfill: bool = False,
    start: str = None,
):
    """
    Checks for inputs to `extend_history`.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe containing categories that are to be extended.
    new_xcat : str
        The name of the new xcat.
    cids : List[str], optional
        The cross sections to extend. If None, all cids available for any category in 'hierarchy' are extended.
    hierarchy : List[str]
         list of categories from best to worst for representation of the concept.
         Inferior categories are only used to extend the history of the superior ones.
         The new category consists of the best representation category values
         and inferior category values that are available prior to any superior.
    backfill : bool, optional
        If True, the new xcat is backfilled to the start date specified by the 'start' parameter.
    start : str, optional
        The start date of the new xcat. If backfill is True, this values will be backfilled up to this date.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df must be a pandas DataFrame")
    if not isinstance(new_xcat, str):
        raise TypeError("new_xcat must be a string")
    if cids is not None:
        if not isinstance(cids, list):
            raise TypeError("cids must be a list")
        if not all(isinstance(cid, str) for cid in cids):
            raise TypeError("cids must be a list of strings")
    if not isinstance(hierarchy, list):
        raise TypeError("hierarchy must be a list")
    if not isinstance(backfill, bool):
        raise TypeError("backfill must be a boolean")
    if start is not None and not isinstance(start, str):
        raise TypeError("start must be a string")
    if not all(isinstance(cat, str) for cat in hierarchy):
        raise TypeError("hierarchy must be a list of strings")

    if backfill and start is None:
        raise ValueError("start must be provided if backfill is True")


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    xcats = ["INFL", "INFL0"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["EUR"] = ["2002-01-01", "2020-09-30", -0.2, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["INFL"] = ["2010-01-01", "2020-09-30", 1, 2, 0.8, 0.5]
    df_xcats.loc["INFL0"] = ["2000-01-01", "2020-09-30", 1, 3, 0.5, 0.2]

    np.random.seed(0)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    df = extend_history(
        dfd, "INFL1", cids, ["INFL", "INFL0"], backfill=False, start=None
    )

    pass



"""
Functions used to visualize correlations across categories or cross-sections of panels.
"""

import itertools
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as sch
import seaborn as sns
from matplotlib import pyplot as plt

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import _map_to_business_day_frequency, reduce_df
from macrosynergy.management.types import QuantamentalDataFrame


def view_correlation(
    df: pd.DataFrame,
    xcats: Union[str, List[str]] = None,
    cids: List[str] = None,
    xcats_secondary: Optional[Union[str, List[str]]] = None,
    cids_secondary: Optional[List[str]] = None,
    start: str = None,
    end: str = None,
    val: str = "value",
    freq: str = None,
    cluster: bool = False,
    lags: dict = None,
    lags_secondary: Optional[dict] = None,
    title: str = None,
    size: Tuple[float] = (14, 8),
    title_fontsize: int = 20,
    fontsize: int = 14,
    max_color: float = None,
    show: bool = True,
    xcat_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    xcat_secondary_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    cid_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    cid_secondary_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    cbar_shrink: Union[float, int] = 0.5,
    cbar_fontsize: int = 12,
    **kwargs: Any,
):
    """
    Displays a correlation matrix across categories or cross-sections of panels.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and at least one column with values of interest.
    xcats : List[str]
        extended categories to be correlated. Default is all in the DataFrame. If xcats
        contains only one category the correlation coefficients across cross sections are
        displayed. If xcats contains more than one category, the correlation coefficients
        across categories are displayed. Additionally, the order of the xcats received will
        be mirrored in the correlation matrix.
    cids : List[str]
        cross sections to be correlated. Default is all in the DataFrame.
    xcats_secondary : List[str]
        an optional second set of extended categories. If xcats_secondary is provided,
        correlations will be calculated between the categories in xcats and xcats_secondary.
    cids_secondary : List[str]
        an optional second list of cross sections. If cids_secondary is provided
        correlations will be calculated and visualized between these two sets.
    start : str
        earliest date in ISO format. Default is None and earliest date in df is used.
    end : str
        latest date in ISO format. Default is None and latest date in df is used.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    freq : str
        frequency option. Per default the correlations are calculated based on the
        native frequency of the datetimes in 'real_date', which is business daily. Down-
        sampling options include weekly ('W'), monthly ('M'), or quarterly ('Q') mean.
    cluster : bool
        if True the series in the correlation matrix are reordered by hierarchical
        clustering. Default is False.
    lags : dict
        optional dictionary of lags applied to respective categories. The key will be
        the category and the value is the lag or lags. If a category has multiple lags
        applied, pass in a list of lag values. The lag factor will be appended to the
        category name in the correlation matrix. If xcats_secondary is not none, this
        parameter will specify lags for the categories in xcats.
    lags_secondary : dict
        optional dictionary of lags applied to the second set of categories if
        xcats_secondary is provided.
    title : str
        chart heading. If none is given, a default title is used.
    size : Tuple[float]
        two-element tuple setting width/height of figure. Default is (14, 8).
    title_fontsize : int
        font size of the title. Default is 20.
    fontsize : int
        font size of the title. Default is 14.
    max_color : float
        maximum values of positive/negative correlation coefficients for color scale.
        Default is none. If a value is given it applies symmetrically to positive and
        negative values.
    show : bool
        if True the figure will be displayed, else the axis object is returned. Default
        is True.
    xcat_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for xcats. A list should be in the same
        order as xcats, a dictionary should map from each xcat to its label.
    xcat_secondary_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for xcats_secondary.
    cid_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for cids. A list should be in the same
        order as cids, a dictionary should map from each cid to its label.
    cid_secondary_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for cids_secondary.
    cbar_shrink : Union[float, int]
        shrinkage factor of the color bar. Default is 0.5.
    cbar_fontsize : int
        font size of the color bar. Default is 12.
    **kwargs : Dict
        Arbitrary keyword arguments that are passed to seaborn.heatmap.


    .. note::
        Lags (`lags`) can include a 0 if the original should also be correlated.

    .. note::
        The function displays the heatmap of a correlation matrix across categories or
        cross-sections (depending on which parameter has received multiple elements).
    """

    df = QuantamentalDataFrame(df[["cid", "xcat", "real_date", val]])

    if freq is not None:
        freq = _map_to_business_day_frequency(freq=freq, valid_freqs=["W", "M", "Q"])

    xcats = xcats if isinstance(xcats, list) else [xcats]

    if max_color is not None:
        assert isinstance(max_color, float), "Parameter max_color must be type <float>."

    if not isinstance(cbar_shrink, (int, float)):
        raise ValueError(
            "The parameter `cbar_shrink` must be of type <int> or <float>."
        )
    if not isinstance(cbar_fontsize, int):
        raise ValueError("The parameter `cbar_fontsize` must be of type <int>.")

    min_color = None if max_color is None else -max_color
    mask = None
    xlabel = ""
    ylabel = ""

    missing_data_msg = (
        "The provided dataframe does not contain any data for the "
        "specified categories: {xcats}. Please check the data."
    )

    xcat_labels = _parse_labels(xcats, xcat_labels, label_type="xcats")
    cid_labels = _parse_labels(cids, cid_labels, label_type="cids")

    # If more than one set of xcats or cids have been supplied.
    if xcats_secondary or cids_secondary:
        if xcats_secondary:
            xcat_secondary_labels = _parse_labels(
                xcats_secondary, xcat_secondary_labels, label_type="xcats_secondary"
            )
            xcats_secondary = (
                xcats_secondary
                if isinstance(xcats_secondary, list)
                else [xcats_secondary]
            )
        else:
            xcats_secondary = xcats
            xcat_secondary_labels = xcat_labels

        if cids_secondary:
            cid_secondary_labels = _parse_labels(
                cids_secondary, cid_secondary_labels, label_type="cids_secondary"
            )
        else:
            cids_secondary = cids
            cid_secondary_labels = cid_labels

        df1, xcats, cids = reduce_df(df.copy(), xcats, cids, start, end, out_all=True)
        df2, xcats_secondary, cids_secondary = reduce_df(
            df.copy(), xcats_secondary, cids_secondary, start, end, out_all=True
        )
        for _df, _xc in zip([df1, df2], [xcats, xcats_secondary]):
            if _df.empty:
                raise ValueError(missing_data_msg.format(xcats=_xc))

        s_date = min(df1["real_date"].min(), df2["real_date"].min()).strftime(
            "%Y-%m-%d"
        )
        e_date = max(df1["real_date"].max(), df2["real_date"].max()).strftime(
            "%Y-%m-%d"
        )

        # If only one xcat, we will compute cross sectional correlation.
        if len(xcats) == 1 and len(xcats_secondary) == 1:
            df_w1 = _transform_df_for_cross_sectional_corr(df=df1, val=val, freq=freq)
            df_w2 = _transform_df_for_cross_sectional_corr(df=df2, val=val, freq=freq)

            if title is None:
                title = (
                    f"Cross-sectional correlation of {xcats[0]} and {xcats_secondary[0]} "
                    f"from {s_date} to "
                    f"{e_date}"
                )
            xlabel = f"{xcat_labels[xcats[0]]} cross-sections"
            ylabel = f"{xcat_secondary_labels[xcats_secondary[0]]} cross-sections"

            new_xcat_labels = xcat_labels
            new_xcat_labels = xcat_labels
            
            df_w1 = df_w1.rename(columns=cid_labels)
            df_w2 = df_w2.rename(columns=cid_secondary_labels)

        # If more than one xcat in at least one set, we will compute cross category
        # correlation.
        else:
            df_w1, new_xcat_labels = _transform_df_for_cross_category_corr(
                df=df1,
                xcats=xcats,
                val=val,
                freq=freq,
                lags=lags,
                labels_dict=xcat_labels,
            )
            df_w2, new_xcat_secondary_labels = _transform_df_for_cross_category_corr(
                df=df2,
                xcats=xcats_secondary,
                val=val,
                freq=freq,
                lags=lags_secondary,
                labels_dict=xcat_secondary_labels,
            )
            df_w1 = df_w1.rename(columns=new_xcat_labels)
            df_w2 = df_w2.rename(columns=new_xcat_secondary_labels)
            if title is None:
                title = f"Cross-category correlation from {s_date} to " f"{e_date}"
        corr = (
            pd.concat([df_w1, df_w2], axis=1, keys=["df_w1", "df_w2"])
            .corr()
            .loc["df_w2", "df_w1"]
        )

        if cluster:
            # Since the correlation matrix is not necessarily symmetric, clustering is
            # done in two stages.
            if corr.shape[0] > 1:
                corr = _cluster_correlations(corr=corr, is_symmetric=False)

            if corr.shape[1] > 1:
                corr = _cluster_correlations(corr=corr.T, is_symmetric=False).T

    # If there is only one set of xcats and cids.
    else:
        df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)
        if df.empty:
            raise ValueError(missing_data_msg.format(xcats=xcats))

        s_date: str = df["real_date"].min().strftime("%Y-%m-%d")
        e_date: str = df["real_date"].max().strftime("%Y-%m-%d")

        if len(xcats) == 1:
            df_w = _transform_df_for_cross_sectional_corr(df=df, val=val, freq=freq, cid_labels=cid_labels)

            if title is None:
                title = (
                    f"Cross-sectional correlation of {xcats[0]} from {s_date} to "
                    f"{e_date}"
                )
            new_xcat_labels = xcat_labels

        else:
            df_w, new_xcat_labels = _transform_df_for_cross_category_corr(
                df=df,
                xcats=xcats,
                val=val,
                freq=freq,
                lags=lags,
                labels_dict=xcat_labels,
            )

            if title is None:
                title = f"Cross-category correlation from {s_date} to {e_date}"

        df_w = df_w.rename(columns=new_xcat_labels)
        corr = df_w.corr(method="pearson")

        if cluster:
            corr = _cluster_correlations(corr=corr, is_symmetric=True)

        # Mask for the upper triangle.
        # Return a copy of an array with the elements below the k-th diagonal zeroed. The
        # mask is implemented because correlation coefficients are symmetric.
        mask: np.ndarray = np.triu(np.ones_like(corr, dtype=bool))

    ax: plt.Axes
    fig, ax = plt.subplots(figsize=size)
    with sns.axes_style("white"):
        with sns.axes_style("ticks"):
            sns.heatmap(
                corr,
                mask=mask,
                cmap="vlag_r",
                center=0,
                vmin=min_color,
                vmax=max_color,
                square=False,
                linewidths=0.5,
                cbar_kws={"shrink": cbar_shrink},
                xticklabels=True,
                yticklabels=True,
                **kwargs,
            )

            cbar = ax.collections[0].colorbar
            cbar.ax.tick_params(labelsize=cbar_fontsize)

    ax.set(xlabel=xlabel, ylabel=ylabel)
    ax.set_title(title, fontsize=title_fontsize)

    ax.tick_params(axis="x", labelsize=fontsize)
    ax.tick_params(axis="y", labelsize=fontsize)

    plt.tight_layout()
    if show:
        plt.show()
        return
    else:
        return ax


def _parse_labels(keys: List[str], labels: Union[List[str], Dict[str, str]], label_type: str):
    """
    Parse labels for cross-sections or categories for correlation plot.

    Parameters
    ----------
    keys : List[str]
        a list of keys for which there may be labels (cids or xcats).
    labels : Union[List[str], Dict[str, str]]
        optional list or dictionary of labels for the keys.
    label_type : str
        Description of the key type for error messages.
    """

    labels_dict = {}
    if labels is not None:
        assert len(labels) == len(keys), (
            f"The number of labels provided for the {label_type} must match the "
            f"number of {label_type}."
        )
        if isinstance(labels, list):
            for key, label in zip(keys, labels):
                labels_dict[key] = label
        elif isinstance(labels, dict):
            labels_dict = labels
        else:
            raise ValueError(f"The labels parameter for {label_type} must be a list or a dictionary.")
    else:
        labels_dict = {key: key for key in keys}
    return labels_dict


def _transform_df_for_cross_sectional_corr(
    df: pd.DataFrame, val: str = "value", freq: str = None, cid_labels: Optional[Dict[str, str]] = None
) -> pd.DataFrame:
    """
    Pivots dataframe and down-samples according to the specified frequency so that
    correlation can be calculated between cross-sections.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized JPMaQS DataFrame.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    freq : str
        Down-sampling frequency option. By default values are not down-sampled.

    Returns
    -------
    pd.Dataframe
        The transformed dataframe.
    """

    df_w = df.pivot(index="real_date", columns="cid", values=val)
    if freq is not None:
        df_w = df_w.resample(freq).mean()

    if cid_labels is not None:
        df_w = df_w.rename(columns=cid_labels)

    return df_w


def _transform_df_for_cross_category_corr(
    df: pd.DataFrame,
    xcats: List[str],
    val: str,
    freq: str = None,
    lags: dict = None,
    labels_dict: Optional[Dict[str, str]] = None,
):
    """
    Pivots dataframe and down-samples according to the specified frequency so that
    correlation can be calculated between extended categories.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized JPMaQS DataFrame.
    xcats : List[str]
        extended categories to be correlated.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    freq : str
        Down-sampling frequency. By default values are not down-sampled.
    lags : dict
        optional dictionary of lags applied to respective categories.

    Returns
    -------
    pd.Dataframe
        The transformed dataframe.
    """
    if labels_dict is not None:
        labels_dict = labels_dict.copy()
    df_w: pd.DataFrame = df.pivot(
        index=("cid", "real_date"), columns="xcat", values=val
    )
    # Down-sample according to the passed frequency.
    if freq is not None:
        df_w = df_w.groupby(
            [pd.Grouper(level="cid"), pd.Grouper(level="real_date", freq=freq)],
            observed=True,
        ).mean()

    # Apply the lag mechanism, to the respective categories, after the down-sampling.
    if lags is not None:
        df_w, xcat_tracker = lag_series(df_w=df_w, lags=lags, xcats=xcats)

        # Order the correlation DataFrame to reflect the order of the categories
        # parameter. Will replace the official category name with the lag appended name.
        order = [
            [x] if x not in xcat_tracker.keys() else xcat_tracker[x] for x in xcats
        ]
        order = list(itertools.chain(*order))
        if labels_dict is not None:
            # If labels_dict is provided, rename each key of labels_dict to be the valu.
            for old_name, new_names in xcat_tracker.items():
                if len(new_names) == 1:
                    labels_dict[new_names[0]] = labels_dict[old_name]
                else:
                    for new_name in new_names:
                        labels_dict[new_name] = (
                            labels_dict[old_name] + f" {new_name.split('_')[-1]}"
                        )
    else:
        order = xcats

    df_w = df_w[order]
    return df_w, labels_dict


def lag_series(
    df_w: pd.DataFrame, lags: dict, xcats: List[str]
) -> Tuple[pd.DataFrame, Dict[str, List[str]]]:
    """
    Method used to lag respective categories.

    Parameters
    ----------
    df_w : ~pandas.DataFrame
        multi-index DataFrame where the columns are the categories, and the two indices
        are the cross-sections and real-dates.
    lags : dict
        dictionary of lags applied to respective categories.
    xcats : List[str]
        extended categories to be correlated.
    """

    lag_type = "The lag data structure must be of type <dict>."
    assert isinstance(lags, dict), lag_type

    lag_xcats = (
        f"The categories referenced in the lagged dictionary must be "
        f"present in the defined DataFrame, {xcats}."
    )
    assert set(lags.keys()).issubset(set(xcats)), lag_xcats

    # Modify the dictionary to adjust for single categories having multiple lags.
    # The respective lags will be held inside a list.
    lag_copy = {}
    xcat_tracker: defaultdict = defaultdict(list)
    for xcat, shift in lags.items():
        if isinstance(shift, int):
            lag_copy[xcat + f"_L{shift}"] = shift
            xcat_tracker[xcat].append(xcat + f"_L{shift}")
        else:
            xcat_temp = [xcat + f"_L{s}" for s in shift]
            # Merge the two dictionaries.
            lag_copy = {**lag_copy, **dict(zip(xcat_temp, shift))}
            xcat_tracker[xcat].extend(xcat_temp)

    df_w_copy = df_w.copy()
    # Handle for multi-index DataFrame. The interior index represents the
    # timestamps.
    for xcat, shift in lag_copy.items():
        category = xcat[:-3]
        clause = isinstance(lags[category], list)
        first_lag = category in df_w.columns

        if clause and not first_lag:
            # Duplicate the column if multiple lags on the same category and the
            # category's first lag has already been implemented. Always access
            # the series from the original DataFrame.
            df_w[xcat] = df_w_copy[category]
        else:
            # Otherwise, modify the name.
            df_w = df_w.rename(columns={category: xcat})
        # Shift the respective column (name will have been adjusted to reflect
        # lag).
        df_w[xcat] = df_w.groupby(level=0)[xcat].shift(shift)

    return df_w, xcat_tracker


def _cluster_correlations(
    corr: pd.DataFrame, is_symmetric: bool = False
) -> pd.DataFrame:
    """
    Rearrange a correlation dataframe so that more similar values are clustered.

    Parameters
    ----------
    corr : ~pandas.DataFrame
        dataframe representing a correlation matrix.
    is_symmetric : bool
        if True, rows and columns are rearranged identically. If False, only rows are
        reordered.

    Returns
    -------
    pd.Dataframe
        The sorted correlation dataframe.
    """

    # Get pairwise distances of the dataframe's rows.
    d = sch.distance.pdist(corr)

    # Perform hierarchical / agglomerative clustering. The clustering method used is
    # Farthest Point Algorithm.
    L = sch.linkage(d, method="complete")

    # The second parameter is the distance threshold, t, which will determine the
    # "number" of clusters. If the distance threshold is too small, none of the data
    # points will form a cluster, so n different clusters are returned.
    # If there are any clusters, the categories contained in the cluster will be
    # adjacent.
    ind = sch.fcluster(L, 0.5 * d.max(), "distance")

    indices = [corr.index.tolist()[i] for i in list((np.argsort(ind)))]

    if is_symmetric:
        corr = corr.loc[indices, indices]
    else:
        corr = corr.loc[indices, :]

    return corr


if __name__ == "__main__":
    np.random.seed(0)

    # Un-clustered correlation matrices.

    cids = ["AUD", "CAD", "GBP", "USD", "NZD", "EUR"]
    cids_dmsc = ["CHF", "NOK", "SEK"]
    cids_dmec = ["DEM", "ESP", "FRF", "ITL", "NLG"]
    cids += cids_dmec
    cids += cids_dmsc
    xcats = ["XR", "CRY", "XR2", "CRY2"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["EUR"] = ["2002-01-01", "2020-09-30", -0.2, 2]
    df_cids.loc["DEM"] = ["2003-01-01", "2020-09-30", -0.3, 2]
    df_cids.loc["ESP"] = ["2003-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["FRF"] = ["2003-01-01", "2020-09-30", -0.2, 2]
    df_cids.loc["ITL"] = ["2004-01-01", "2020-09-30", -0.2, 0.5]
    df_cids.loc["NLG"] = ["2003-01-01", "2020-12-30", -0.1, 0.5]
    df_cids.loc["CHF"] = ["2003-01-01", "2020-12-30", -0.3, 2.5]
    df_cids.loc["NOK"] = ["2010-01-01", "2020-12-30", -0.1, 0.5]
    df_cids.loc["SEK"] = ["2010-01-01", "2020-09-30", -0.1, 0.5]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2010-01-01", "2020-10-30", 1, 2, 0.95, 0.5]
    df_xcats.loc["XR2",] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY2",] = ["2010-01-01", "2020-10-30", 1, 2, 0.95, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    start = "2012-01-01"
    end = "2020-09-30"

    lag_dict = {"XR": [0, 2, 5]}
    xcat_labels = {
        "XR": "Excess returns",
        "CRY": "Carry",
        "XR2": "Excess returns 2",
        "CRY2": "Carry 2",
    }
    # Clustered correlation matrices. Test hierarchical clustering.
    view_correlation(
        df=dfd,
        xcats=["XR", "CRY", "XR2", "CRY2"],
        xcats_secondary=["CRY", "XR", "CRY2", "XR2"],
        cids=cids,
        cids_secondary=cids[:4],
        start=start,
        end=end,
        val="value",
        freq=None,
        cluster=True,
        title="Correlation Matrix",
        size=(14, 8),
        max_color=None,
        lags=None,
        lags_secondary=None,
        annot=True,
        fmt=".2f",
        xcat_labels=xcat_labels,
        xcat_secondary_labels={
            "XR": "Excess returns",
            "CRY": "Carry",
            "XR2": "Excess returns 2",
            "CRY2": "Carry 2",
        },
    )
    print(xcat_labels)
    
    view_correlation(
        df=dfd,
        xcats=["XR"],
        xcats_secondary=["CRY", "XR"],
        cids=cids,
        cids_secondary=cids[:1],
        start=start,
        end=end,
        val="value",
        freq=None,
        cluster=True,
        title="Correlation Matrix",
        size=(14, 8),
        max_color=None,
        lags=None,
        lags_secondary=None,
        annot=True,
        fmt=".2f",
        xcat_labels={
            "XR": "Excess returns",
        },
        xcat_secondary_labels={
            "CRY": "Carry",
            "XR": "Excess returns",
        },
        # cid_secondary_labels={
        #     "AUD": "Australian Dollar",
        #     "CAD": "Canadian Dollar",
        #     "GBP": "British Pound",
        #     "USD": "US Dollar",
        # }
    )
    
    view_correlation(
        df=dfd,
        xcats=["XR"],
        # xcats_secondary=["CRY", "XR"],
        cids=cids[:4],
        cids_secondary=cids[:1],
        start=start,
        end=end,
        val="value",
        freq=None,
        cluster=True,
        title="Correlation Matrix",
        size=(14, 8),
        max_color=None,
        lags=None,
        lags_secondary=None,
        annot=True,
        fmt=".2f",
    )



import numpy as np
import pandas as pd

from sklearn.base import ClassifierMixin, BaseEstimator
from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils.multiclass import check_classification_targets
class KNNClassifier(ClassifierMixin, BaseEstimator):
    def __init__(self, n_neighbors="sqrt", weights="uniform"):
        """
        Nearest neighbors classifier.

        Parameters
        ----------
        n_neighbors : int, float, or str
            Number of neighbors to use. If int, the number of neighbors to use.
            If float, the fraction of the number of samples to use. If "sqrt",
            the square root of the number of samples is used.
        weights : str
            Weight function used to aggregate neighbors. Possible values are "uniform"
            and "distance".

        Notes
        -----
        The class is a wrapper around the KNeighborsClassifier from scikit-learn. It has
        been implemented to allow for the use of a fraction of the number of samples as
        the number of neighbors to use. In addition, the square root of the number of
        samples is a common rule of thumb for the number of neighbors to use - we wanted
        to allow for this option in cross-validation.
        """
        self._check_init_params(n_neighbors, weights)

        self.n_neighbors = n_neighbors
        self.weights = weights
        self.knn_ = None
        self.classes_ = [-1, 1]

    def fit(self, X, y):
        """
        Fit method.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            The input feature matrix.
        y : pd.Series or np.ndarray
            The target variable.

        Returns
        -------
        self
            The fitted model.
        """
        self._check_fit_params(X, y)

        if self.n_neighbors == "sqrt":
            n = int(np.sqrt(len(X)))
        elif isinstance(self.n_neighbors, float):
            n = int(self.n_neighbors * len(X))
        else:
            n = self.n_neighbors
        self.knn_ = KNeighborsClassifier(n_neighbors=n, weights=self.weights).fit(X, y)

        return self

    def predict(self, X):
        """
        Predict method.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            The input feature matrix.

        Returns
        -------
        np.ndarray
            The predicted values.
        """
        self._check_predict_params(X)

        return self.knn_.predict(X)

    def predict_proba(self, X):
        """
        Predict probability method.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            The input feature matrix.

        Returns
        -------
        np.ndarray
            The predicted probabilities.
        """
        self._check_predict_params(X)

        return self.knn_.predict_proba(X)
    
    def __getattr__(self, attr):
        """
        Return the class attributes.

        Parameters
        ----------
        attr : str
            The attribute to return.

        Returns
        -------
        Any
            The attribute.
        """
        try:
            return getattr(self.knn_, attr)
        except AttributeError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{attr}'"
            )
    
    def _check_init_params(self, n_neighbors, weights):
        """
        Check the parameters passed to the __init__ method.
        """
        # n_neighbors
        if not isinstance(n_neighbors, (int, float, str)):
            raise TypeError("n_neighbors must be an int, float, or str")
        if isinstance(n_neighbors, str) and n_neighbors != "sqrt":
            raise ValueError('n_neighbors must be "sqrt" if it is a str')
        if isinstance(n_neighbors, float) and ((n_neighbors <= 0) or (n_neighbors >= 1)):
            raise ValueError("n_neighbors must be between 0 and 1 if it is a float")
        if isinstance(n_neighbors, int) and n_neighbors <= 0:
            raise ValueError("n_neighbors must be greater than 0 if it is an int")
        if not isinstance(weights, str):
            raise TypeError("weights must be a str")
        if weights not in ["uniform", "distance"]:
            raise ValueError('weights must be "uniform" or "distance')
        
    def _check_fit_params(self, X, y):
        """
        Check the parameters passed to the fit method.
        """
        # Type checks
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError("X must be a pd.DataFrame or np.ndarray")
        if not isinstance(y, (pd.Series, np.ndarray)):
            raise TypeError("y must be a pd.Series or np.ndarray")
        # Value checks 
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of samples")
        if isinstance(X, (pd.DataFrame, pd.Series)) and isinstance(y, pd.Series):
            if not X.index.equals(y.index):
                raise ValueError("X and y must have the same index")
        # Value checks for X
        if isinstance(X, pd.DataFrame):
            if not isinstance(X.index, pd.MultiIndex):
                raise ValueError("X must have a multi-index")
            if len(X.index.levels) != 2:
                raise ValueError("X must have a multi-index with two levels")
            if not X.index.get_level_values(0).dtype == "object":
                raise TypeError("The outer index of X must be strings.")
            if not X.index.get_level_values(1).dtype == "datetime64[ns]":
                raise TypeError("The inner index of X must be datetime.date.")
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for nearest neighbor forecasts is a "
                    "numpy array, it must have two dimensions."
                )
        # Value checks for y
        if isinstance(y, pd.Series):
            if not isinstance(y.index, pd.MultiIndex):
                raise ValueError("y must have a multi-index")
            if len(y.index.levels) != 2:
                raise ValueError("y must have a multi-index with two levels")
            if not y.index.get_level_values(0).dtype == "object":
                raise TypeError("The outer index of y must be strings.")
            if not y.index.get_level_values(1).dtype == "datetime64[ns]":
                raise TypeError("The inner index of y must be datetime.date.")
        if isinstance(y, np.ndarray):
            if y.ndim != 1:
                raise ValueError(
                    "When the target variable for nearest neighbor forecasts is a numpy "
                    "array, it must have one dimension."
                )
        check_classification_targets(y)

    def _check_predict_params(self, X):
        """
        Check the parameters passed to the predict method.
        """
        # Type checks
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError("X must be a pd.DataFrame or np.ndarray")
        # Value checks for X
        if isinstance(X, pd.DataFrame):
            if not isinstance(X.index, pd.MultiIndex):
                raise ValueError("X must have a multi-index")
            if len(X.index.levels) != 2:
                raise ValueError("X must have a multi-index with two levels")
            if not X.index.get_level_values(0).dtype == "object":
                raise TypeError("The outer index of X must be strings.")
            if not X.index.get_level_values(1).dtype == "datetime64[ns]":
                raise TypeError("The inner index of X must be datetime.date.")
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for nearest neighbor forecasts is a "
                    "numpy array, it must have two dimensions."
                )
        
    
if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm
    from macrosynergy.learning import SignalOptimizer, ExpandingKFoldPanelSplit

    from sklearn.linear_model import LinearRegression
    from sklearn.model_selection import GridSearchCV
    from sklearn.metrics import make_scorer, r2_score

    # Randomly generate an unbalanced panel dataset, multi-indexed by cross-section and
    # real_date

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 0]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 0, 1, -0.9, 0]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", 0, 1, 0.8, 0]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = np.sign(dfd["XR"])

    # Fit nearest neighbors classifier
    knn = KNNClassifier(n_neighbors=0.1)
    knn.fit(X, y)
    print(knn.predict(X))



'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.3.4

version: 1.3.4

full_version: 1.3.4

git_revision: dc0c75e9e1e0122bf839af0ecb9d5ffee0d45a4f

release: True

'''
short_version: str = '1.3.4'
version: str = '1.3.4'
full_version: str = '1.3.4'
git_revision: str = 'dc0c75e9e1e0122bf839af0ecb9d5ffee0d45a4f'
release: bool = True
if not release:
    version = full_version



import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin

class ProbabilityEstimator(BaseEstimator, MetaEstimatorMixin, ClassifierMixin):
    """
    Meta estimator to create trading signals based on the probability of going long.

    Parameters
    ----------
    classifier : ClassifierMixin
        A scikit-learn classifier.

    Notes
    -----
    This class stores feature importances as the feature importances of the base estimator
    as well as defining a create_signal method that returns the probability of going long
    in excess of 0.5. This is taken into account when used in the SignalOptimizer class
    in this package.
    """
    def __init__(self, classifier):
        if not isinstance(classifier, ClassifierMixin):
            raise TypeError("classifier must be a scikit-learn classifier.")
        
        self.classifier = classifier
        self.classes_ = [-1,1]

    def fit(self, X, y):
        """
        Fit the underlying classifier.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Pandas dataframe or numpy array of input features.
        y : pd.Series or pd.DataFrame or np.ndarray
            Pandas series, dataframe or numpy array of targets associated with each sample
            in X.
        """
        # Checks
        self._check_fit_params(X, y)

        # Model fitting
        self.classifier.fit(X, y)

        # Store feature importances
        if hasattr(self.classifier, "feature_importances_"):
            self.feature_importances_ = self.classifier.feature_importances_
        elif hasattr(self.classifier, "coef_"):
            self.feature_importances_ = np.abs(self.classifier.coef_) / np.sum(np.abs(self.classifier.coef_))

        return self
    
    def predict(self, X):
        """
        Predict the class labels for the provided data.

        Parameters
        ----------
        X : pd.DataFrame or numpy array
            Input feature matrix.

        Returns
        -------
        y_pred : np.ndarray
            Numpy array of predictions.
        """
        # Checks
        self._check_predict_params(X)

        # Predict
        return self.classifier.predict(X)

    def create_signal(self, X):
        """
        Create a trading signal based on the probability of going long.

        Parameters
        ----------
        X : pd.DataFrame or numpy array
            Input feature matrix.

        Returns
        -------
        y_pred : np.ndarray
            Numpy array of signals.
        """
        # Checks
        self._check_predict_params(X)

        # Create signal
        return self.classifier.predict_proba(X)[:,1] - 0.5
    
    def _check_fit_params(
        self,
        X,
        y,
    ):
        """
        Checks for fit method parameters.
        """
        # X
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for the probability estimator must be either a pandas "
                "dataframe or numpy array."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for the probability estimator is a "
                    "numpy array, it must have two dimensions."
                )
        # y
        if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Target vector for the probability estimator must be either a pandas series, "
                "dataframe or numpy array."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The dependent variable dataframe must have only one column. If used "
                    "as part of an sklearn pipeline, ensure that previous steps return "
                    "a pandas series or dataframe."
                )
        if isinstance(y, np.ndarray):
            if y.ndim != 1:
                raise ValueError(
                    "When the target vector for the probability estimator is a numpy "
                    "array, it must have one dimension."
                )
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the target vector."
            )
        
    def _check_predict_params(self, X):
        """
        Checks for predict method parameters.
        """
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for the probability estimator must be either a pandas "
                "dataframe or numpy array. If used as part of an sklearn pipeline, ensure "
                "that previous steps return a pandas dataframe or numpy array."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for the probability estimator is a "
                    "numpy array, it must have two dimensions. If used as part of an "
                    "sklearn pipeline, ensure that previous steps return a two-dimensional "
                    "data structure."
                )

        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "The number of features in the input feature matrix must match the number "
                "seen in training."
            )
        
    def __getattr__(self, attr):
        """
        Get attributes from the underlying classifier.
        """
        if hasattr(self.classifier, attr):
            return getattr(self.classifier, attr)
        
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")
    
if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf
    import pandas as pd
    import numpy as np

    from sklearn.linear_model import LogisticRegression

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    # Regressor
    X_train = train.drop(columns=["XR"])
    y_train = np.sign(train["XR"])

    pe = ProbabilityEstimator(LogisticRegression()).fit(X_train, y_train)
    print(pe.create_signal(X_train))


from macrosynergy.pnl.naive_pnl import NaivePnL, create_results_dataframe
from macrosynergy.pnl.multi_pnl import MultiPnL

from macrosynergy.pnl.contract_signals import contract_signals
from macrosynergy.pnl.notional_positions import notional_positions
from macrosynergy.pnl.historic_portfolio_volatility import historic_portfolio_vol
from macrosynergy.pnl.proxy_pnl_calc import proxy_pnl_calc
from macrosynergy.pnl.transaction_costs import TransactionCosts
from macrosynergy.pnl.proxy_pnl import ProxyPnL

__all__ = [
    "NaivePnL",
    "MultiPnL",
    "create_results_dataframe",
    "contract_signals",
    "notional_positions",
    "historic_portfolio_vol",
    "proxy_pnl_calc",
    "TransactionCosts",
    "ProxyPnL",
]



"""
Module implementing the `make_relative_category()` function, which calculates indicator
values relative to a set of categories. Designed with sectoral equity returns in mind,
this function can be extended to other data types with a similar structure.
Unlike `make_relative_value()`, which computes values relative to a cross-sectional mean
for a given category, `make_relative_category()` returns values relative to an average
across categories.
"""

import pandas as pd
from typing import List, Set

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.types import QuantamentalDataFrame
import warnings


def make_relative_category(
    df: pd.DataFrame,
    xcats: List[str] = None,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    basket: List[str] = None,
    complete_set: bool = False,
    rel_meth: str = "subtract",
    rel_xcats: List[str] = None,
    postfix: str = "RC",
):
    """
    For each cross section specified in the panel, relative values are calculated by
    either subtracting or dividing by the average across categories at each `real_date`.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and 'value'.
    xcats : List[str]
        all extended categories for which relative values are to be calculated. The user
        must provide the set of categories to be used in the calculation.
    cids : List[str]
        cross sections for which relative values are calculated.
    start : str
        earliest date in ISO format. Default is None and earliest date for which the
        respective category is available is used.
    end : str
        latest date in ISO format. Default is None and latest date for which the
        respective category is available is used.
    blacklist : dict
        cross sections with date ranges that should be excluded from the output.
    basket : List[str]
        categories to be used for the relative value benchmark. The default is every
        category in the DataFrame that is available over the time-period.
        However, the basket parameter can specify a subset of the available
        categories.
    complete_set : bool
        boolean parameter that determines whether each cross-section is required to have
        the full set of extended categories in the basket for a relative value calculation
        to occur. If set to True, the cross-section will be excluded from the output if
        some desired categories are missing. Default is False. If False, the mean, for the
        relative value, will use the subset of categories available for that cross-section.
    rel_meth : str
        method for calculating relative value. Default is 'subtract'. Alternative is
        'divide'.
    rel_xcats : List[str]
        extended category name of the relative values. Will displace the original
        category names: xcat + postfix. The order should reflect the order of the passed
        categories.
    postfix : str
        acronym to be appended to 'xcat' string to give the name for relative value
        category. Only applies if rel_xcats is None. Default is 'RC' relative category

    Returns
    -------
    ~pandas.DataFrame
        standardized DataFrame with the relative values, with the columns:
        'cid', 'xcat', 'real_date' and 'value'.
    """

    col_names = ["cid", "xcat", "real_date", "value"]
    operations = {
        "divide": pd.Series.div,
        "subtract": pd.Series.sub,
    }
    if rel_meth not in operations:
        raise ValueError("rel_meth must be 'subtract' or 'divide'")

    if not isinstance(xcats, (list, str)):
        raise TypeError("xcats must be a list of strings or a single string.")

    if isinstance(xcats, str):
        xcats = [xcats]

    if rel_xcats is not None:
        if not (
            isinstance(rel_xcats, list) and all([isinstance(x, str) for x in rel_xcats])
        ):
            raise ValueError("List of strings expected for `rel_xcats`.")

        if len(xcats) != len(rel_xcats):
            raise ValueError(
                "`rel_xcats` must have the same number of elements as `xcats`."
            )

        rel_xcats_dict = dict(zip(xcats, rel_xcats))
    else:
        rel_xcats_dict = {x: x + postfix for x in xcats}

    df = QuantamentalDataFrame(df)
    # Intersect parameter set to False. Therefore, cross sections across the categories can vary.
    if basket:
        all_xcats: List[str] = list(set(xcats).union(set(basket)))
    else:
        all_xcats = xcats
    if len(all_xcats) < 1:
        all_xcats = None
    dfx = reduce_df(df, all_xcats, cids, start, end, blacklist, out_all=False)

    if cids is None:
        # All cross sections available - union across categories.
        cids = list(dfx["cid"].unique())

    if basket is not None:
        # Basket must be a subset of the available xcats
        miss: Set = set(basket) - set(df["xcat"])
        error_basket = f"The category basket elements {miss} are not specified or are not available."
        if not len(miss) == 0:
            raise ValueError(error_basket)
    else:
        # Default basket is all available cross sections.
        basket = xcats

    available_cids = dfx["cid"].unique()

    if len(xcats) == len(basket) == 1:
        run_error = (
            "Computing the relative value on a single category using a "
            "basket consisting exclusively of the aforementioned category "
            "is an incorrect usage of the function."
        )
        raise ValueError(run_error)

    df_list: List[pd.DataFrame] = []

    # Each cid could have an incomplete basket, we are allowing it to be flexible
    basket_full = sorted(set(basket))
    for i, cid in enumerate(available_cids):
        df_cid = dfx[dfx["cid"] == cid]
        xcats_avl = df_cid["xcat"].unique()

        df_cid: pd.DataFrame = df_cid[["xcat", "real_date", "value"]]

        dfb, basket = _prepare_category_basket(
            df=df_cid,
            cid=cid,
            basket=basket_full,
            xcats_avl=xcats_avl,
            complete_set=complete_set,
        )

        if len(basket) > 0:
            bm = dfb.groupby(by="real_date").mean(numeric_only=True).reset_index()
        else:
            continue

        dfa = df_cid.copy()
        dfa["count"] = dfa.groupby("real_date")["value"].transform("count")
        dfa = dfa.loc[dfa["count"] > 1]

        # The time-index will be delimited by the respective category.
        dfa = pd.merge(dfa, bm, how="left", on=["real_date"], suffixes=["", "_bm"])
        dfa = dfa.sort_values(by=["xcat", "real_date"])

        dfa["value"] = operations[rel_meth](dfa["value"], dfa["value_bm"], axis=0)

        # cleaning
        df_new = dfa.drop(columns=["value_bm", "count"])
        df_new["xcat"] = df_new["xcat"].map(rel_xcats_dict).astype("category")
        df_new["cid"] = pd.Categorical.from_codes(
            codes=[0] * len(df_new), categories=[cid]
        )

        if (
            df_new.sort_values(["xcat", "real_date"])[col_names].isna().all().all()
            or (len(df_new) == 0)
            or df_new.empty
        ):
            continue

        df_list.append(df_new.sort_values(["xcat", "real_date"])[col_names])

    if len(df_list) == 0:
        raise ValueError(
            "No relative category DataFrame could be created. "
            "Check the input parameters and the DataFrame."
        )

    return QuantamentalDataFrame(
        pd.concat(df_list).reset_index(drop=True),
        categorical=df.InitializedAsCategorical,
    )


def _prepare_category_basket(
    df: pd.DataFrame,
    cid: str,
    basket: List[str],
    xcats_avl: List[str],
    complete_set: bool,
):
    """
    Relative cid-specific indicators can be defined over different sets of categories.
    We will determine the respective basket given the available categories for the
    respective cross section.

    Parameters
    ----------
    df : ~pandas.DataFrame
        long JPMaQS DataFrame of single category.
    cid : str
        target cross section for the relative value calculation.
    basket : ~pandas.DataFrame
        set of categories to be used for the relative value benchmark if available.
    xcats_avl : List[str]
        categories available for the target cross section.
    complete_set : bool
        if True, the basket is only calculated if all categories, held in the basket,
        are available for that respective category.

    Returns
    -------
    ~pandas.DataFrame
        DataFrame with the categories available for the target cross section.
    List[str]
        List of categories available for the target cross section.
    """

    xcats_used: List[str] = sorted(set(basket) & set(xcats_avl))
    xcats_miss: List[str] = list(set(basket) - set(xcats_used))

    # if any missing cids, warn
    if xcats_miss:
        if complete_set:
            xcats_used = []
            err_str: str = (
                f"The cross section, {cid}, is missing {xcats_miss} which are included "
                f"in the basket {basket}. Therefore, the cross section will be excluded "
                "from the returned DataFrame."
            )
            print(err_str)
            warnings.warn(err_str, UserWarning)
        else:
            err_str: str = (
                f"The cross section, {cid}, is missing {xcats_miss} from the requested "
                f"basket. The new basket will be {xcats_used}."
            )
            print(err_str)
            warnings.warn(err_str, UserWarning)

    # Reduce the DataFrame to the specified basket given the available cross sections.
    dfb = df.loc[df["xcat"].isin(xcats_used)]

    return dfb, xcats_used


if __name__ == "__main__":
    # Simulate DataFrame.

    cids = ["AUD", "CAD", "GBP", "NZD"]
    xcats = ["XR1", "XR2", "CRY1", "CRY2"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD"] = ["2001-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2002-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR1"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["XR2"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["CRY1"] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["CRY2"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Simulate blacklist
    black = {"AUD": ["2000-01-01", "2003-12-31"], "GBP": ["2018-01-01", "2100-01-01"]}

    # Applications
    dfd_relative_xr = make_relative_category(
        dfd,
        xcats=["XR1", "XR2"],
        cids=None,
        blacklist=None,
        rel_meth="subtract",
        rel_xcats=None,
    )

    dfd_1_black = make_relative_category(
        dfd,
        xcats=["CRY1", "CRY2"],
        cids=None,
        blacklist=black,
        rel_meth="divide",
        rel_xcats=None,
        postfix="_RELRATIO",
    )



import numpy as np 
import pandas as pd

from sklearn.base import BaseEstimator, RegressorMixin
from scipy.optimize import minimize

class GlobalLocalRegression(BaseEstimator, RegressorMixin):
    r"""
    Linear panel model with hierarchical shrinkage of country-specific (local) coefficients 
    towards unknown global coefficients. Learning means that both country-specific and 
    global coefficients are estimated from data.

    Parameters
    ----------
    local_lambda : float, default=1
        Regularization strength to pull local coefficients towards global coefficients.
    global_lambda : float, default=1
        Regularization strength to pull global coefficients towards zero. 
    positive : bool, default=False
        Whether to constrain all coefficients to be positive. Default is False.
    fit_intercept : bool, default=True
        Whether to fit an intercept term. Default is True.
    min_xs_samples : int, default=36
        Minimum number of samples required in each group for the group to be considered 
        a contribution to the mean squared error component of the loss function.

    Notes
    -----
    A panel can be modelled from a global perspective, where time series of all countries
    are "pooled" or stacked together, meaning that samples from different countries are 
    treated as independent. This is called a pooled regression. With one model fit on all
    countries' data, this is a high-bias, low-variance model. 

    Alternatively, country-by-country regressions can be fit, with a separate model for
    each country. This is low-bias but high-variance, since each model sees less data.

    This implies that a balance can be found between these two extremes by balancing 
    this bias-variance trade-off. Introduction of bias to the country-by-country models
    can lead to a potentially substantial reduction in variance. Mathematically, this fit 
    is found by minimizing the sum of squared residuals for each country, with a term 
    that penalizes deviation of country-specific coefficients from a global coefficient. 
    The global coefficient is also penalized to prevent it from growing too large.

    The loss function is as follows:

    .. math::

            L(\{\beta_i\}_{i=1}^{C}, \beta) = \frac{1}{C} \sum_{i = 1}^{C} \left [ \frac{1}{n_{i}}  \sum_{t=1}^{n_{i}} (y_{it} - x_{it}^{\intercal} \beta_{i})^2 \right ] + \lambda_{\text{local}} \sum_{i=1}^{C} ||\beta_i - \beta||_{2}^{2} + \lambda_{\text{global}} ||\beta||_{2}^{2}
    """
    def __init__(self, local_lambda = 1, global_lambda = 1, positive = False, fit_intercept = True, min_xs_samples = 36):
        # Attributes
        self.local_lambda = local_lambda
        self.global_lambda = global_lambda
        self.positive = positive
        self.fit_intercept = fit_intercept
        self.min_xs_samples = min_xs_samples
    
    def fit(self, X, y, sample_weight = None):
        """
        Fit the global-local model.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix, multi-indexed by `cid` and `real_date`.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X, multi-indexed by `cid` and
            `real_date`.
        sample_weight : np.ndarray, optional
            Sample weights for each sample in X. If provided, it should be a 1D
            array with the same length as the number of samples in X. If None, all samples
            are treated equally.

        Returns
        -------
        self
            Fitted estimator.
        """
        # Fit
        if self.fit_intercept:
            X = X.copy()
            X.insert(0, "intercept", 1)
            
        self.cids_ = sorted(X.index.get_level_values(0).unique())
        self.n_cids_ = len(self.cids_)
        self.n_features_ = X.shape[1]
        self.X_cid_ = {}
        self.y_cid_ = {}
        self.Xy_cid_weights_ = None if sample_weight is None else {}
        
        for cid in self.cids_:
            self.X_cid_[cid] = X.loc[cid].values
            self.y_cid_[cid] = y.loc[cid].values
            if sample_weight is not None:
                self.Xy_cid_weights_[cid] = sample_weight[y.index.get_level_values(0)==cid]

        # Initialise with zeros
        x0 = np.zeros((self.n_cids_ + 1) * self.n_features_)

        # Optional bounds
        if self.fit_intercept:
            if self.positive:
                # Place positive restrictions on the feature weights not intercepts
                bounds = []
                for _ in range(self.n_cids_ + 1):
                    bounds.append((None, None)) 
                    bounds.extend([(0, None)] * (self.n_features_ - 1)) 
            else:
                bounds = [(None, None)] * len(x0)
        else:
            if self.positive:
                bounds = [(0, None)] * len(x0)
            else:
                bounds = [(None, None)] * len(x0)

        result = minimize(
            fun=self.loss,
            x0=x0,
            method="L-BFGS-B",
            bounds=bounds,
            jac=self.loss_derivative,
        )

        coefs = result.x.reshape(self.n_cids_ + 1, self.n_features_)
        self.cid_coefs_ = {cid: coefs[i] for i, cid in enumerate(self.cids_)}
        
        if self.fit_intercept:
            self.coef_ = coefs[-1, 1:]
            self.intercept_ = coefs[-1, 0]
        else:
            self.coef_ = coefs[-1]
            self.intercept_ = None

        return self
    
    def loss(self, weights):
        """
        Loss function for the global-local regression model.

        Parameters
        ----------
        weights : np.ndarray
            Flattened array of weights, where the last `n_features_` elements correspond
            with the global coefficients and the rest correspond to the local coefficients
            for each country.
        """
        weights = weights.reshape(self.n_cids_ + 1, self.n_features_)
        total_loss = 0.0
        global_beta = weights[-1, :]
        
        # Likelihood
        for i, g in enumerate(self.cids_):
            X_g = self.X_cid_[g]
            y_g = self.y_cid_[g]
            preds = X_g @ weights[i]
            residuals = y_g - preds

            if self.Xy_cid_weights_ is not None:
                w_g = self.Xy_cid_weights_[g]
                total_loss += np.average(residuals**2, weights=w_g)
            else:
                total_loss += np.mean(residuals**2)
        
        # Local-to-global regularization term
        if self.local_lambda > 0:
            reg = np.sum((weights[:-1] - global_beta)**2)
            total_loss += self.local_lambda * reg
            
        # Global regularization term
        if self.global_lambda > 0:
            reg = np.sum(global_beta**2)
            total_loss += self.global_lambda * reg
            
        return total_loss
    
    def loss_derivative(self, weights):
        """
        Derivative of the loss function with respect to the weights.

        Parameters
        ----------
        weights : np.ndarray
            Flattened array of weights, where the last `n_features_` elements correspond
            with the global coefficients and the rest correspond to the local coefficients
            for each country. 
        """
        weights = weights.reshape(self.n_cids_ + 1, self.n_features_)
        global_beta = weights[-1]
        
        grads = np.zeros_like(weights)

        for i, cid in enumerate(self.cids_):
            X_i = self.X_cid_[cid]
            y_i = self.y_cid_[cid]
            beta_i = weights[i]
            preds = X_i @ beta_i
            residual = preds - y_i
            
            if self.Xy_cid_weights_ is not None:
                w_i = self.Xy_cid_weights_[cid]
                grad_i = 2 * (X_i.T @ (residual * w_i)) / np.sum(w_i)
            else:
                grad_i = 2 * X_i.T @ residual / len(residual)

            grad_i += 2 * self.local_lambda * (beta_i - global_beta)
            grads[i] = grad_i

        # Gradient for global beta
        grad_g = 2 * self.local_lambda * np.sum(global_beta - weights[:-1], axis=0) + 2 * self.global_lambda * global_beta
        grads[-1] = grad_g

        return grads.flatten()
    
    def predict(self, X):
        """
        Predict the target values for the given input data.

        Parameters
        ----------
        X : pd.DataFrame
            Input features for prediction.

        Returns
        -------
        np.ndarray
            Predicted target values.
        """
        if self.fit_intercept:
            X = X.copy()
            X.insert(0, "intercept", 1)
            
        group_indices = X.index.get_level_values(0)
        X_values = X.values
        preds = np.zeros(len(X))

        for group in self.cids_:
            mask = group_indices == group
            if mask.any():
                beta = self.cid_coefs_.get(group, np.zeros(self.n_features_))
                preds[mask] = X_values[mask] @ beta

        return preds
    
if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    # Regressor
    X_train = train.drop(columns=["XR"])
    y_train = np.sign(train["XR"])

    glr = GlobalLocalRegression().fit(X_train, y_train)
    print(glr.predict(X_train))


from .base_splitters import BasePanelSplit, KFoldPanelSplit, WalkForwardPanelSplit
from .kfold_splitters import ExpandingKFoldPanelSplit, RollingKFoldPanelSplit, RecencyKFoldPanelSplit
from .walk_forward_splitters import ExpandingIncrementPanelSplit, ExpandingFrequencyPanelSplit

__all__ = [
    'BasePanelSplit',
    'KFoldPanelSplit',
    'WalkForwardPanelSplit',
    # KFold splitters
    'ExpandingKFoldPanelSplit',
    'RollingKFoldPanelSplit',
    'RecencyKFoldPanelSplit',
    # Walk-forward splitters
    'ExpandingIncrementPanelSplit',
    'ExpandingFrequencyPanelSplit',
]


"""
Functionality to visualize time series data as line charts.
"""

import numpy as np
import pandas as pd
from typing import Dict, List, Tuple, Optional, Union

from macrosynergy.management.simulate import make_qdf

import macrosynergy.visuals as msv


def view_timelines(
    df: pd.DataFrame,
    xcats: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    intersect: bool = False,
    val: str = "value",
    cumsum: bool = False,
    start: str = None,
    end: Optional[str] = None,
    ncol: int = 3,
    legend_ncol: int = 1,
    same_y: bool = True,
    all_xticks: bool = False,
    xcat_grid: bool = False,
    xcat_labels: Union[Optional[List[str]], Dict] = None,
    cid_labels: Union[Optional[List[str]], Dict] = None,
    sort_cid_labels: bool = False,
    single_chart: bool = False,
    label_adj: float = 0.05,
    title: Optional[str] = None,
    title_adj: float = 1.005,
    title_xadj: float = 0.5,
    title_fontsize: int = 18,
    cs_mean: bool = False,
    size: Tuple[float, float] = (12, 7),
    aspect: float = 1.618,
    height: float = 2.85,
    legend_fontsize: int = 12,
    blacklist: Dict = None,
    ax_hline: Optional[Dict] = 0.0,
):
    """
    Displays a grid with subplots of time line charts of one or more categories.

    Parameters
    ----------
    df : ~pandas.Dataframe
        standardized DataFrame with the necessary columns: 'cid', 'xcat', 'real_date'
        and at least one column with values of interest.
    xcats : List[str]
        extended categories to plot. Default is all in DataFrame.
    cids : List[str]
        cross sections to plot. Default is all in DataFrame. If this contains only one
        cross section a single line chart is created.
    intersect : bool
        if True only retains cross-sections that are available for all categories.
        Default is False.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    cumsum : bool
        plot cumulative sum of the values over time. Default is False.
    start : str
        earliest date in ISO format. Default is earliest date available.
    end : str
        latest date in ISO format. Default is latest date available.
    ncol : int
        number of columns in facet grid. Default is 3.
    legend_ncol : int
        number of columns in legend. Default is 1.
    same_y : bool
        if True (default) all plots in facet grid share same y axis.
    all_xticks : bool
        if True x-axis tick labels are added to all plots in grid. Default is False, i.e
        only the lowest row displays the labels.
    xcat_grid : bool
        if True, shows a facet grid of line charts for each category for a single cross
        section. Default is False, only one cross section is allowed with this option.
    xcat_labels : Union[Optional[List[str]], Dict]
        labels to be used for xcats. If not defined, the labels will be identical to
        extended categories.
    cid_labels : Union[Optional[List[str]], Dict]
        labels to be used for cids. If not defined, the labels will be identical to
        cross-sections.
    sort_cid_labels : bool
        if True, sorts the cross-sectional labels in the grid alphabetically. Otherwise,
        the order of `cids` is preserved. Default is False.
    single_chart : bool
        if True, all lines are plotted in a single chart.
    title : str
        chart heading. Default is no title.
    title_adj : float
        parameter that sets top of figure to accommodate title. Default is 0.95.
    title_xadj : float
        parameter that sets x position of title. Default is 0.5.
    title_fontsize : int
        font size of title. Default is 16.
    label_adj : float
        parameter that sets bottom of figure to fit the label. Default is 0.05.
    cs_mean : bool
        if True this adds a line of cross-sectional averages to the line charts. This is
        only allowed for function calls with a single category. Default is False.
    size : Tuple[float]
        two-element tuple setting width/height of single cross section plot. Default is
        (12, 7). This is irrelevant for facet grid.
    aspect : float
        width-height ratio for plots in facet. Default is 1.7.
    height : float
        height of plots in facet. Default is 3.
    legend_fontsize : int
        font size of legend. Default is 12.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the dataframe.
    ax_hline : Union[float, Dict]
        if float, this value is used for all cross-sections to draw a horizontal line.
        horizontal line to be drawn at this value. Default is 0.0. This can be a dict
        with keys as cross-sections and values as the value of the horizontal line. 
        If none is provided, no horizontal line is drawn.
    """

    msv.timelines(
        df=df,
        xcats=xcats,
        cids=cids,
        intersect=intersect,
        val=val,
        cumsum=cumsum,
        start=start,
        end=end,
        ncol=ncol,
        same_y=same_y,
        all_xticks=all_xticks,
        xcat_grid=xcat_grid,
        xcat_labels=xcat_labels,
        cid_labels=cid_labels,
        sort_cid_labels=sort_cid_labels,
        single_chart=single_chart,
        title=title,
        legend_ncol=legend_ncol,
        cs_mean=cs_mean,
        title_fontsize=title_fontsize,
        legend_fontsize=legend_fontsize,
        # HC Params
        aspect=aspect,
        title_adj=title_adj,
        height=height,
        size=size,
        label_adj=label_adj,
        title_xadj=title_xadj,
        blacklist=blacklist,
        ax_hline=ax_hline,
    )


if __name__ == "__main__":
    np.random.seed(42)
    cids = ["AUD", "CAD", "GBP", "NZD"]
    xcats = ["XR", "CRY", "INFL", "FXXR"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.2, 0.2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2012-01-01", "2020-09-30", -0.1, 3]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )

    df_xcats.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["INFL"] = ["2015-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2013-01-01", "2020-10-30", 1, 2, 0.95, 0.5]
    df_xcats.loc["FXXR"] = ["2013-01-01", "2020-10-30", 1, 2, 0.95, 0.5]

    dfd: pd.DataFrame = make_qdf(df_cids, df_xcats, back_ar=0.75)
    # sort by cid, xcat, and real_date
    dfd = dfd.sort_values(["cid", "xcat", "real_date"])
    ctr: int = -1
    for xcat in xcats[:2]:
        for cid in cids[:2]:
            ctr *= -1
            mask = (dfd["cid"] == cid) & (dfd["xcat"] == xcat)
            dfd.loc[mask, "value"] = (
                10
                * ctr
                * np.arange(dfd.loc[mask, "value"].shape[0])
                / (dfd.loc[mask, "value"].shape[0] - 1)
            )

    dfdx = dfd[~((dfd["cid"] == "AUD") & (dfd["xcat"] == "XR"))]

    view_timelines(
        dfd,
        xcats=["XR", "CRY"],
        cids=cids[0],
        size=(10, 5),
        title="AUD Return and Carry",
        aspect=3,
        ax_hline=0
    )

    view_timelines(
        dfd,
        xcats=["XR", "CRY", "INFL"],
        cids=cids[0],
        xcat_grid=True,
        title_adj=0.8,
        xcat_labels=["Return", "Carry", "Inflation"],
        title="AUD Return, Carry & Inflation",
        aspect=3,
    )

    view_timelines(dfd, xcats=["CRY"], cids=cids, ncol=4, title="Carry", cs_mean=True)

    view_timelines(
        dfd, xcats=["XR"], cids=cids[:2], ncol=2, cumsum=True, same_y=False, aspect=2
    )

    dfd = dfd.set_index("real_date")
    view_timelines(
        dfd,
        xcats=["XR"],
        cids=['AUD', 'NZD', 'CAD'],
        ncol=2,
        cumsum=True,
        same_y=False,
        aspect=2,
        single_chart=False,
        cid_labels={
            'NZD': 'New Zealand Dollar',
            'CAD': 'Canadian Dollar',
            'AUD': 'Australian Dollar'
        },
        sort_cid_labels=False,
    )



"""
Contains mathematical utility functions used across the package.
"""

import itertools
from numbers import Number
from typing import List

import numpy as np
import pandas as pd

from macrosynergy.management.simulate import make_qdf


def expanding_mean_with_nan(
    dfw: pd.DataFrame, absolute: bool = False
) -> List[np.float64]:
    """
    Calculate the expanding mean of a DataFrame's values across rows, handling NaN values.

    This function computes the expanding (cumulative) mean of all elements in the 
    DataFrame `dfw`, row-by-row. NaN values are ignored in the summation, ensuring they 
    do not affect the calculation. If `absolute` is set to True, it uses the absolute 
    values of elements for the expanding mean calculation. The function returns a list 
    of expanding mean values, with each element corresponding to the expanding mean up to 
    that row.

    Parameters
    ----------
    dfw : pd.DataFrame
        A DataFrame with a datetime index (or convertible to datetime) and numeric data 
        across its columns. The index is expected to represent timestamps.
    absolute : bool, optional
        If True, computes the expanding mean using the absolute values of the DataFrame's 
        elements, by default False.

    Returns
    -------
    List[np.float64]
        A list containing the expanding mean for each row of the DataFrame.

    Raises
    ------
    TypeError
        If `dfw` is not a DataFrame, if its index cannot be converted to timestamps, or if 
        `absolute` is not a boolean.
    """

    if not isinstance(dfw, pd.DataFrame):
        raise TypeError("Method expects to receive a pd.DataFrame.")

    # cast the index to pd.Timestamp, if error raise TypeError
    try:
        dfw.index = pd.to_datetime(dfw.index)
    except:
        raise TypeError("The index of the DataFrame must be of type `<pd.Timestamp>`.")

    if not isinstance(absolute, bool):
        raise TypeError("The parameter `absolute` must be of type `<bool>`.")

    data: np.ndarray = dfw.to_numpy()

    no_rows: int = dfw.shape[0]
    no_columns: int = dfw.shape[1]
    no_elements: int = no_rows * no_columns

    one_dimension_arr: np.ndarray = data.reshape(no_elements)
    if absolute:
        one_dimension_arr = np.absolute(one_dimension_arr)

    rolling_summation: List[float] = [
        np.nansum(one_dimension_arr[0 : (no_columns * i)])
        for i in range(1, no_rows + 1)
    ]

    # Determine the number of active cross-sections per timestamp. Required for computing
    # the rolling mean.
    data_arr = data.astype(dtype=np.float32)
    # Sum across the rows.
    active_cross = np.sum(~np.isnan(data_arr), axis=1)
    rolling_active_cross = list(itertools.accumulate(active_cross))

    mean_calc = lambda m, d: m / d
    ret = list(map(mean_calc, rolling_summation, rolling_active_cross))

    return np.array(ret)


def ewm_sum(df: pd.DataFrame, halflife: Number):
    """
    Compute the exponentially weighted moving sum of a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame in the wide format for which to calculate weights.
    halflife : Number
        The halflife of the exponential decay.
    """

    if not isinstance(df, pd.DataFrame):
        raise TypeError("Method expects to receive a pd.DataFrame.")
    if not isinstance(halflife, Number):
        raise TypeError("The parameter `halflife` must be of type `<Number>`.")

    weights = calculate_cumulative_weights(df, halflife)
    return df.ewm(halflife=halflife).mean().mul(weights, axis=0)


def calculate_cumulative_weights(df: pd.DataFrame, halflife: Number):
    """
    Calculate the cumulative moving exponential weights for a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame in the wide format for which to calculate weights.
    halflife : Number
        The halflife of the exponential decay.
    """

    n = len(df)
    raw_weights = [(1 / 2) ** (i / halflife) for i in range(n)]

    cumulative_weights = np.cumsum(raw_weights)

    return pd.Series(cumulative_weights, index=df.index)


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    # Define the cross-sections over different timestamps such that the pivoted DataFrame
    # includes NaN values: more realistic testcase.
    df_cids.loc["AUD"] = ["2022-01-01", "2022-02-01", 0.5, 2]
    df_cids.loc["CAD"] = ["2022-01-10", "2022-02-01", 0.5, 2]
    df_cids.loc["GBP"] = ["2022-01-20", "2022-02-01", -0.2, 0.5]
    df_cids.loc["USD"] = ["2022-01-01", "2022-02-01", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2022-01-05", "2022-02-01", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR"] = ["2010-01-01", "2022-02-01", 0, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2011-01-01", "2022-02-01", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2022-02-01", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2013-01-01", "2022-02-01", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd_xr = dfd[dfd["xcat"] == "XR"]

    dfw = dfd_xr.pivot(index="real_date", columns="cid", values="value")
    no_rows = dfw.shape[0]

    ret_mean = expanding_mean_with_nan(dfw)



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin


class NaiveRegressor(BaseEstimator, RegressorMixin):
    """
    Equally weighted unbiased factor model.

    Notes
    -----
    Given a collection of factors that are theoretically positively correlated with a
    dependent variable, a plausible signal is a simple average of those factors. This is
    effectively a linear regression model with zero intercept and equal weights for all
    factors.

    This is a useful benchmark model which works well when the factors are as
    uncorrelated as possible with one another, because it offers a layer of
    diversification on the underlying return drivers. When the user has strong priors,
    this is often a competitive model that is difficult to beat.

    However, it is vital for the features to have been preprocessed to have a positive
    theoretical correlation with the target variable.
    """

    def fit(self, X, y=None):
        """
        Fit method.

        Parameters
        ----------
        X : pd.DataFrame, pd.Series or np.ndarray
            The input feature matrix.
        y : pd.DataFrame, pd.Series or np.ndarray
            The target variable.

        Returns
        -------
        self
            The fitted model.
            
        Notes
        -----
        This method involves fully trusting one's priors and thus requires no learning
        element. As a consequence, no training set information is needed.
        """
        # Checks
        if not isinstance(X, (pd.DataFrame, pd.Series, np.ndarray)):
            raise TypeError(
                "X must be a pandas DataFrame, pandas Series or numpy array"
            )
        elif isinstance(X, np.ndarray) and ((X.ndim > 2) or (X.ndim < 1)):
            raise ValueError(
                "When X is a numpy array, it must have either 1 or 2 dimensions."
            )

        # No learning needed since priors are fully trusted
        self.n = len(X)
        if isinstance(X, pd.Series):
            self.p = 1
        elif isinstance(X, np.ndarray):
            self.p = X.shape[1] if X.ndim == 2 else 1
        else:
            self.p = X.shape[1]

        self.X_type = type(X)

        return self

    def predict(self, X):
        """
        Predict method.

        Notes
        -----
        The predictions are simply the average of the features across columns of the input
        feature matrix.
        """
        # Checks
        if not isinstance(X, (pd.DataFrame, pd.Series, np.ndarray)):
            raise TypeError(
                "X must be a pandas DataFrame, pandas Series or numpy array"
            )
        elif isinstance(X, np.ndarray) and ((X.ndim > 2) or (X.ndim < 1)):
            raise ValueError(
                "When X is a numpy array, it must have either 1 or 2 dimensions."
            )
        if not isinstance(X, self.X_type):
            raise ValueError(
                "X must be of the same type as the input matrix to the fit() method."
            )

        if isinstance(X, np.ndarray) and X.ndim == 1:
            p = 1
        elif isinstance(X, pd.Series):
            p = 1
        else:
            p = X.shape[1]
        if p != self.p:
            raise ValueError(
                "X must have the same number of columns as the input matrix to the "
                "fit() method."
            )

        # Return the naive signal
        if isinstance(X, pd.DataFrame):
            return np.mean(X.values, axis=1)
        elif isinstance(X, pd.Series):
            return X.values
        elif isinstance(X, np.ndarray) and X.ndim == 1:
            return X
        else:
            return np.mean(X, axis=1)


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2012-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2012-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2012-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    Xy = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, freq="M", lag=1, xcat_aggs=["last", "sum"]
    ).dropna()
    X = Xy.iloc[:, :-1]
    y = Xy.iloc[:, -1]

    naive = NaiveRegressor()
    naive.fit(X, y)
    print(naive.predict(X))



import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, RegressorMixin
from scipy.optimize import minimize

import numbers
import warnings
from functools import partial
from sklearn.exceptions import ConvergenceWarning


class LADRegressor(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        alpha=0,
        shrinkage_type="l1",
        tol=None,
        maxiter=None,
    ):
        r"""
        Linear regression with L1 loss.

        Parameters
        ----------
        fit_intercept : bool, default=True
            Whether or not to add an intercept to the model.
        positive: bool, default=False
            Whether or not to enforce non-negativity of model weights,
            with exception to a model intercept.
        alpha: float, default=0
            Shrinkage hyperparameter.
        shrinkage_type: str, default="l1"
            Type of shrinkage regularization to perform.
        tol: float, default=None
            Tolerance for convergence of the learning algorithm (SLSQP).
            This is passed into the 'tol' parameter of the scipy.optimize.minimize
            function.
        maxiter: int, default=None
            Maximum number of iterations for the learning algorithm (SLSQP).
            This is passed into the 'maxiter' key of the options dictionary in the
            scipy.optimize.minimize function.

        Notes
        -----
        A dependent variable is modelled as a linear combination of the input features.
        The weights associated with each feature (and the intercept) are determined by
        finding the weights that minimise the average absolute model residuals.

        If `alpha` is positive, then shrinkage-based regularization is applied to the
        non-intercept model coefficients. The type of shrinkage is determined by the
        `shrinkage_type` parameter. If `shrinkage_type` is "l1", then L1 regularization
        is applied. If `shrinkage_type` is "l2", then L2 regularization is applied.

        Mathematically, the following optimization problem is solved:

        .. math::

        \underset{w, b}{\text{argmin}} \frac{1}{N} \sum_{i=1}^{N} |y_{i} - x_{i}^{T}w - b|
        + 2N \times \alpha \times L_{p}(w)^{p}

        where:
            - N is the number of samples in the training data.
            - :math:`y_{i}` is the response for the i-th sample.
            - :math:`x_{i}` is the feature vector for the i-th sample.
            - :math:`w` is the vector of model coefficients.
            - :math:`b` is the model intercept.
            - :math:`L_{p}(w)` is the :math:`L_{p}` norm of the model coefficients, where
                :math:`p` is 1 or 2.
            - :math:`alpha` is the regularization hyperparameter.
        """
        # Checks
        self._check_init_params(
            fit_intercept, positive, alpha, shrinkage_type, tol, maxiter
        )

        # Initialise
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.alpha = alpha
        self.shrinkage_type = shrinkage_type
        self.tol = tol
        self.maxiter = maxiter

        # Set values of quantities to learn to None
        self.coef_ = None
        self.intercept_ = None

    def fit(
        self,
        X,
        y,
        sample_weight: np.ndarray = None,
    ):
        """
        Learn LAD regression model parameters.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Input feature matrix.
        y : pd.Series or pd.DataFrame or np.ndarray
            Target vector associated with each sample in X.
        sample_weight : np.ndarray, default=None
            Numpy array of sample weights to create a weighted LAD regression model.
        """
        # Checks
        self._check_fit_params(X, y, sample_weight)

        # Fit
        X = X.copy()
        y = y.copy()

        if self.fit_intercept:
            if isinstance(X, pd.DataFrame):
                X.insert(0, "intercept", 1)
            else:
                X = np.insert(X, 0, 1, axis=1)

        n_cols = X.shape[1]

        # Optimization bounds
        if self.positive:
            if self.fit_intercept:
                bounds = [(None, None)] + [(0, None)] * (n_cols - 1)
            else:
                bounds = [(0, None)] * n_cols
        else:
            bounds = [(None, None)] * n_cols

        # Set initial weights
        init_weights = (
            np.zeros(n_cols) if not self.fit_intercept else np.zeros(n_cols - 1)
        )
        if self.fit_intercept:
            init_intercept = np.mean(y)
            init_weights = np.concatenate(([init_intercept], init_weights))

        X = X if isinstance(X, np.ndarray) else X.values
        y = y.squeeze() if isinstance(y, np.ndarray) else y.values.squeeze()

        optim_results = minimize(
            fun=partial(
                self._l1_loss,
                X=X,
                y=y,
                sample_weight=sample_weight,
                alpha=self.alpha,
                shrinkage_type=self.shrinkage_type,
            ),
            x0=init_weights,
            method="SLSQP",
            bounds=bounds,
            tol=self.tol,
            options={"maxiter": self.maxiter} if self.maxiter else None,
        )

        # Handle optimization results
        if not optim_results.success:
            warnings.warn(
                "LAD regression failed to converge. Try increasing the number of "
                "iterations or decreasing the tolerance level. The "
                "scipy.optimize.minimize message is: {}".format(optim_results.message),
                ConvergenceWarning,
            )

            return self

        if self.fit_intercept:
            self.intercept_ = optim_results.x[0]
            self.coef_ = optim_results.x[1:]
        else:
            self.intercept_ = 0.0
            self.coef_ = optim_results.x

        return self

    def predict(self, X):
        """
        Predict dependent variable using the fitted LAD regression model.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Input feature matrix.

        Returns
        -------
        y_pred : np.ndarray
            Numpy array of predictions.

        Notes
        -----
        If the model learning algorithm failed to converge, the predict method will return
        an array of zeros. This has the interpretation of no buy/sell signal being
        triggered based on this model.
        """
        # Checks
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for the LADRegressor must be either a pandas "
                "dataframe or numpy array. If used as part of an sklearn pipeline, ensure "
                "that previous steps return a pandas dataframe or numpy array."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for LADRegressor forecasts is a numpy "
                    "array, it must have two dimensions. If used as part of an sklearn "
                    "pipeline, ensure that previous steps return a two-dimensional data "
                    "structure."
                )

        if X.shape[1] != len(self.coef_):
            raise ValueError(
                "The number of features in the input feature matrix must match the number "
                "of model coefficients."
            )

        if isinstance(X, pd.DataFrame):
            if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
                raise ValueError(
                    "All columns in the input feature matrix must be numeric."
                )
            if X.isnull().values.any():
                raise ValueError(
                    "The input feature matrix for the LADRegressor must not contain any "
                    "missing values."
                )
        else:
            if not np.issubdtype(X.dtype, np.number):
                raise ValueError(
                    "All elements in the input feature matrix must be numeric."
                )
            if np.isnan(X).any():
                raise ValueError(
                    "The input feature matrix for the LADRegressor must not contain any "
                    "missing values."
                )

        # Predict
        if self.coef_ is None:
            return np.zeros(X.shape[0])

        if self.fit_intercept:
            return (X.dot(self.coef_) + self.intercept_).values
        else:
            return X.dot(self.coef_).values

    def _l1_loss(
        self,
        weights,
        X,
        y,
        sample_weight=None,
        alpha: float = 0,
        shrinkage_type: str = "l1",
    ):
        """
        Determine L1 loss induced by 'weights'.

        Parameters
        ----------
        weights : np.ndarray
            LADRegressor model coefficients to be optimised.
        X : np.ndarray
            Input features.
        y : np.ndarray
            Targets associated with each sample in X.
        sample_weight : np.ndarray, default=None
            Sample weights to create a weighted LAD regression model.
        alpha : float, default=0
            Shrinkage hyperparameter.
        shrinkage_type : str, default="l1"
            Type of shrinkage regularization to perform.

        Returns
        -------
        loss : float
            L1 loss induced by 'weights'.
        """
        if sample_weight is None:
            sample_weight = np.ones(X.shape[0])

        raw_residuals = y - X @ weights
        abs_residuals = np.abs(raw_residuals)
        weighted_abs_residuals = abs_residuals * sample_weight

        if alpha > 0:
            if shrinkage_type == "l1":
                if self.fit_intercept:
                    l1_norm = np.sum(np.abs(weights[1:]))
                else:
                    l1_norm = np.sum(np.abs(weights))
                return (
                    np.mean(weighted_abs_residuals) + 2 * X.shape[0] * alpha * l1_norm
                )
            elif shrinkage_type == "l2":
                if self.fit_intercept:
                    l2_norm = np.sum(weights[1:] ** 2)
                else:
                    l2_norm = np.sum(weights**2)
                return (
                    np.mean(weighted_abs_residuals) + 2 * X.shape[0] * alpha * l2_norm
                )

        return np.mean(weighted_abs_residuals)

    def _check_init_params(
        self, fit_intercept, positive, alpha, shrinkage_type, tol, maxiter
    ):
        """
        Checks for constructor parameters.
        """
        # fit_intercept
        if not isinstance(fit_intercept, bool):
            raise TypeError("The fit_intercept parameter must be a boolean.")
        # positive
        if not isinstance(positive, bool):
            raise TypeError("The positive parameter must be a boolean.")
        # alpha
        if (not isinstance(alpha, numbers.Real)) or isinstance(alpha, bool):
            raise TypeError("The alpha parameter must be numeric.")
        if alpha < 0:
            raise ValueError("The alpha parameter must be non-negative.")
        # shrinkage_type
        if not isinstance(shrinkage_type, str):
            raise TypeError("The shrinkage_type parameter must be a string.")
        if shrinkage_type not in ["l1", "l2"]:
            raise ValueError(
                "The shrinkage_type parameter must be either 'l1' or 'l2'."
            )
        # tol
        if tol is not None:
            if (not isinstance(tol, numbers.Real)) or isinstance(tol, bool):
                raise TypeError("The tol parameter must be numeric.")
            if tol < 0:
                raise ValueError("The tol parameter must be non-negative.")
        # maxiter
        if maxiter is not None:
            if (not isinstance(maxiter, numbers.Real)) or isinstance(maxiter, bool):
                raise TypeError("The maxiter parameter must be an integer.")
            if maxiter <= 0:
                raise ValueError("The maxiter parameter must be positive.")

    def _check_fit_params(self, X, y, sample_weight):
        """
        Checks for fit method parameters.
        """
        # Check type of X
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for the LADRegressor must be a pandas dataframe or "
                "numpy array."
            )
        # Check structure of X: no missing values and all numeric
        if isinstance(X, pd.DataFrame):
            if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
                raise ValueError(
                    "All columns in the input feature matrix must be numeric."
                )
            if X.isnull().values.any():
                raise ValueError(
                    "The input feature matrix for the LADRegressor must not contain any "
                    "missing values."
                )
        else:
            if not np.issubdtype(X.dtype, np.number):
                raise ValueError(
                    "All elements in the input feature matrix must be numeric."
                )
            if np.isnan(X).any():
                raise ValueError(
                    "The input feature matrix for the LADRegressor must not contain any "
                    "missing values."
                )

        # Check type of y
        if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Dependent variable for the LADRegressor must be a pandas series, "
                "dataframe or numpy array."
            )
        # Check structure of y: single numeric column, no missing values
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The dependent variable dataframe must have only one column. If used "
                    "as part of an sklearn pipeline, ensure that previous steps return "
                    "a pandas series or dataframe."
                )
            if not pd.api.types.is_numeric_dtype(y.iloc[:, 0]):
                raise ValueError(
                    "The dependent variable dataframe must contain only numeric values."
                )
            if y.isnull().values.any():
                raise ValueError(
                    "The dependent variable dataframe must not contain any missing values."
                )
        elif isinstance(y, pd.Series):
            if not pd.api.types.is_numeric_dtype(y):
                raise ValueError(
                    "The dependent variable series must contain only numeric values."
                )
            if y.isnull().values.any():
                raise ValueError(
                    "The dependent variable series must not contain any missing values."
                )
        else:
            if y.ndim != 1:
                raise ValueError(
                    "The dependent variable numpy array must be 1D. If the dependent "
                    "variable is 2D, please either flatten the array or double check the "
                    "contents of `y`."
                )
            if not np.issubdtype(y.dtype, np.number):
                raise ValueError(
                    "All elements in the dependent variable vector must be numeric."
                )
            if np.isnan(y).any():
                raise ValueError(
                    "The dependent variable for the LADRegressor must not contain any "
                    "missing values."
                )

        if len(X) != len(y):
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the dependent variable."
            )

        # sample_weight
        if sample_weight is not None:
            if not isinstance(sample_weight, np.ndarray):
                try:
                    sample_weight = sample_weight.to_numpy().flatten()
                except Exception as e:
                    try:
                        sample_weight = np.array(sample_weight).flatten()
                    except Exception as e:
                        raise TypeError(
                            "The sample weights must be contained within a numpy array."
                        )
            if sample_weight.ndim != 1:
                raise ValueError("The sample weights must be a 1D numpy array.")
            for w in sample_weight:
                if not isinstance(w, numbers.Number):
                    raise ValueError(
                        "All elements of the sample weights must be numeric."
                    )
            if len(sample_weight) != X.shape[0]:
                raise ValueError(
                    "The number of sample weights must match the number of samples in the "
                    "input feature matrix."
                )


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    # Fit model
    model = LADRegressor(
        fit_intercept=True, positive=False, alpha=1, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")



"""
A subclass inheriting from `macrosynergy.visuals.plotter.Plotter`, designed to render a
facet plot containing any number of subplots. Given that the class allows returning a
`matplotlib.pyplot.Figure`, one can easily add any number of subplots, even the
FacetPlot itself: effectively allowing for a recursive facet plot.
"""

from numbers import Number
from typing import Any, Dict, List, Optional, Tuple, Union, Callable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.gridspec import GridSpec
from statsmodels.graphics.tsaplots import plot_acf

from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.visuals.plotter import Plotter


def _get_square_grid(
    num_plots: int,
) -> Tuple[int, int]:
    """
    Given the number of plots, return a tuple of grid dimensions that is closest to a
    square grid.

    Parameters
    ----------
    num_plots : int
        Number of plots.

    Returns
    -------
    Tuple[int, int]
        Tuple of grid dimensions.
    """

    sqrt_num_plots: float = np.ceil(np.sqrt(num_plots))
    grid_dim: Tuple[int, int] = (int(sqrt_num_plots), int(sqrt_num_plots))
    gd_copy: Tuple[int, int] = grid_dim
    # the number of plots is less than grid_dim[0] * grid_dim[1],
    # so iteratively try and reduce the row and column dimensions until sweet spot is
    # found.
    while 0 != 1:
        if gd_copy[0] < gd_copy[1]:
            gd_copy: Tuple[int, int] = (gd_copy[0], gd_copy[1] - 1)
        else:
            gd_copy: Tuple[int, int] = (gd_copy[0] - 1, gd_copy[1])

        # if gd_copy[0] * gd_copy[1] is more than num_plots, copy to grid_dim. if smaller,
        # break.
        if gd_copy[0] * gd_copy[1] >= num_plots:
            grid_dim: Tuple[int, int] = gd_copy
        else:
            break

    return grid_dim


class FacetPlot(Plotter):
    """
    Class for rendering a facet plot containing any number of subplots. Inherits from
    `macrosynergy.visuals.plotter.Plotter`.

    Parameters
    ----------
    df : ~pandas.DataFrame
        A DataFrame with the following columns: 'cid', 'xcat', 'real_date', and at least
        one metric from - 'value', 'grading', 'eop_lag', or 'mop_lag'.
    cids : List[str]
        A list of cids to select from the DataFrame. If None, all cids are selected.
    xcats : List[str]
        A list of xcats to select from the DataFrame. If None, all xcats are selected.
    metrics : List[str]
        A list of metrics to select from the DataFrame. If None, all metrics are
        selected.
    intersect : bool
        if True only retains cids that are available for all xcats. Default is False.
    tickers : List[str]
        A list of tickers to select from the DataFrame. If None, all tickers are
        selected.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the data frame. If
        one cross-section has several blacklist periods append numbers to the cross-section
        code.
    start : str
        ISO-8601 formatted date string. Select data from this date onwards. If None, all
        dates are selected.
    end : str
        ISO-8601 formatted date string. Select data up to and including this date. If
        None, all dates are selected.
    """

    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        intersect: bool = False,
        tickers: Optional[List[str]] = None,
        blacklist: Optional[Dict[str, List[str]]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            intersect=intersect,
            tickers=tickers,
            blacklist=blacklist,
            start=start,
            end=end,
            *args,
            **kwargs,
        )

    def _get_grid_dim(
        self,
        tickers: List[str],
        ncols: Optional[int] = None,
        attempt_square: bool = False,
        cid_grid: bool = False,
        xcat_grid: bool = False,
        cid_xcat_grid: bool = False,
        _cids: Optional[List[str]] = None,
        _xcats: Optional[List[str]] = None,
    ) -> Tuple[int, int]:
        """
        Returns a tuple of grid dimensions that matches the plot settings.
        """
        # only one of cid_grid, xcat_grid, cid_xcat_grid or attempt_square can be True
        if sum([cid_grid, xcat_grid, cid_xcat_grid]) > 1:
            raise ValueError(
                "Only one of `cid_grid`, `xcat_grid`, or "
                "`cid_xcat_grid` can be True."
            )

        if _cids is None:
            _cids: List[str] = self.cids
        if _xcats is None:
            _xcats: List[str] = self.xcats

        if attempt_square:
            target_var: str = tickers
            if cid_grid:
                target_var: List[str] = _cids
            elif xcat_grid:
                target_var: List[str] = _xcats
            return _get_square_grid(num_plots=len(target_var))

        if cid_grid or xcat_grid:
            tks: List[str] = _cids if cid_grid else _xcats
            return self._get_grid_dim(
                tickers=tks,
                ncols=ncols,
                attempt_square=attempt_square,
                _cids=_cids,
                _xcats=_xcats,
            )

        if cid_xcat_grid:
            return (len(_xcats), len(_cids))

        if ncols is not None:
            return (
                ncols,
                int(np.ceil(len(tickers) / ncols)),
            )

        raise ValueError("Unable to infer grid dimensions.")

    def lineplot(
        self,
        # plot arguments
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metric: Optional[str] = None,
        ncols: int = 3,
        attempt_square: bool = False,
        cid_grid: bool = False,
        xcat_grid: bool = False,
        cid_xcat_grid: bool = False,
        grid_dim: Optional[Tuple[int, int]] = None,
        compare_series: Optional[str] = None,
        share_y: bool = False,
        share_x: bool = False,
        interpolate: bool = False,
        # xcats_mean: bool = False,
        # title arguments
        figsize: Tuple[Number, Number] = (16.0, 9.0),
        title: Optional[str] = None,
        title_fontsize: int = 22,
        title_xadjust: Optional[Number] = None,
        title_yadjust: Optional[Number] = None,
        # subplot axis arguments
        ax_grid: bool = False,
        ax_hline: Optional[Number] = 0.0,
        ax_vline: Optional[str] = None,
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 12,
        # subplot arguments
        facet_size: Optional[Tuple[Number, Number]] = None,
        facet_titles: Optional[Union[List[str], Dict[str, str]]] = None,
        facet_title_fontsize: int = 14,
        facet_title_xadjust: Number = 0.5,
        facet_title_yadjust: Number = 1.0,
        facet_xlabel: Optional[str] = None,
        facet_ylabel: Optional[str] = None,
        # legend arguments
        legend: bool = True,
        legend_labels: Optional[List[str]] = None,
        legend_loc: Optional[str] = "lower center",
        legend_fontsize: int = 12,
        legend_ncol: int = 1,
        legend_bbox_to_anchor: Optional[Tuple[Number, Number]] = None,  # (1.0, 0.5),
        legend_frame: bool = True,
        # return args
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        plot_func=None,
        plot_func_kwargs: Dict[str, Any] = {},
        *args,
        **kwargs,
    ):
        """
        Showing a FacetPlot composed of linear plots from the data available in the
        `FacetPlot` object after initialization. Passing any of the arguments used to
        initialize the `FacetPlot` object will cause the object to be re-initialized
        with the new arguments, and the plot will be rendered from the new object state.

        Parameters
        ----------
        cids : List[str]
            A list of `cids` to select from the DataFrame. If None, the `cids` are selected
            from the object initialization, or all if were specified.
        xcats : List[str]
            A list of `xcats` to select from the DataFrame. If None, all `xcats` are selected
            from the object initialization, or all if were specified.
        metric : str
            the metric to plot. Default is the first metric in the DataFrame.
        ncols : int
            number of columns in the grid. Default is 3.
        attempt_square : bool
            attempt to make the facet grid square. Ignores `ncols` when `True`. Default
            is `False`.
        cid_grid : bool
            Create the facet grid such that each facet is a `cid`.
        xcat_grid : bool
            Create the facet grid such that each facet is an `xcat`.
        cid_xcat_grid : bool
            Create the facet grid such that each facet is an individual ticker. Each
            "row" contains plots for the same `cid`, and each "column" would contain plots
            for the same `xcat`. Therefore, this mode does not respect the `ncols` or
            `attempt_square` arguments.
        grid_dim : Tuple[int, int]
            A tuple of integers specifying the number of rows and columns in the grid.
        compare_series : str
            Used with `cid_grid` with a single `xcat`. If specified, the series
            specified will be plotted in each facet, as a red dashed line. This is useful
            for comparing a single series, such as a benchmark/average. Ensure that the
            comparison series is in the dataframe, and not filtered out when initializing
            the `FacetPlot` object. Default is `None`.
        share_y : bool
            whether to share the y-axis across all plots. Default is `True`.
        share_x : bool
            whether to share the x-axis across all plots. Default is `True`.
        interpolate : bool
            if `True`, gaps in the time series will be interpolated. Default is `False`.
        figsize : Tuple[Number, Number]
            a tuple of floats specifying the width and height of the figure. Default is
            `(16.0, 9.0)`.
        title : str
            the title of the plot. Default is `None`.
        title_fontsize : int
            the font size of the title. Default is `20`.
        title_xadjust : float
            the x-adjustment of the title. Default is `None`.
        title_yadjust : float
            the y-adjustment of the title. Default is `None`.
        ax_grid : bool
            whether to show the grid on the axes, applied to all plots. Default is
            `True`.
        ax_hline : Number
            the value of the horizontal line on the axes, applied to all plots. Default
            is `None`, meaning no horizontal line will be shown.
        ax_vline : str
            the value of the vertical line on the axes, applied to all plots. The value
            must be a ISO-8601 formatted date-string. Default is `None`, meaning no vertical
            line will be shown.
        x_axis_label : str
            the label for the x-axis. Default is `None`.
        y_axis_label : str
            the label for the y-axis. Default is `None`.
        axis_fontsize : int
            the font size of the axis labels. Default is `12`.
        facet_size : Tuple[Number, Number]
            a tuple of floats specifying the width and height of each facet. Default is
            `None`, meaning the facet size will be inferred from the `figsize` argument. If
            specified, the `figsize` argument will be ignored and the figure size will be
            inferred from the dimensions of the facet grid and the facet size.
        facet_titles : List[str]
            a list of strings specifying the titles of each facet. Default is `None`,
            meaning all facets will have the full `ticker`, `cid`, or `xcat` as the title.
            If no `facet_titles` are required, pass an empty list - `facet_titles=[]`.
        facet_title_fontsize : int
            the font size of the facet titles. Default is `12`.
        facet_title_xadjust : float
            the x-adjustment of the facet titles. Default is `None`.
        facet_title_yadjust : float
            the y-adjustment of the facet titles. Default is `None`.
        facet_xlabel : bool
            The label to be used as the axis-label/title for the x-axis of each facet.
            Default is `None`, meaning no label will be shown.
        facet_ylabel : bool
            The label to be used as the axis-label/title for the y-axis of each facet.
            Default is `None`, meaning no label will be shown.
        facet_label_fontsize : int
            the font size of the facet labels. Default is `12`.
        legend : bool
            Show the legend. Default is `True`. When using `cid_xcat_grid`, the legend
            will not be shown as it is redundant.
        legend_labels : list
            Labels for the legend. Default is `None`, meaning a list identifying the
            various `cids`/`xcats` will be used.
        legend_loc : str
            Location of the legend. Default is `center left`. See
            https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html for more
            information.
        legend_fontsize : int
            Font size of the legend. Default is `12`.
        legend_ncol : int
            Number of columns in the legend. Default is `1`.
        legend_bbox_to_anchor : tuple
            Bounding box for the legend. Default is `(1.0, 0.5)`.
        legend_frame : bool
            Show the legend frame. Default is `True`.
        show : bool
            Show the plot. Default is `True`.
        save_to_file : str
            Save the plot to a file. Default is `None`.
        dpi : int
            DPI of the saved image. Default is `300`.
        return_figure : bool
            Return the figure object. Default is `False`.
        plot_func : callable
            The plotting function to use. The function must have parameters real_dates, values,
            and ax, and it will be passed these.Default is None.
        plot_func_kwargs : dict
            Additional arguments to pass to the plotting function. Default is `{}`.


        .. note::
            `compare_series` can only be used when the series is not removed by
            `reduce_df()` in the object initialization.

        .. note::
            `facet_titles` and `legend` are overridden in this mode.
        """

        comp_series_flag: bool = False

        if compare_series:
            if compare_series not in set(
                (kwargs["df"] if "df" in kwargs else self.df)[["cid", "xcat"]]
                .drop_duplicates()
                .apply(lambda x: "_".join(x), axis=1)
                .tolist()
            ):
                comp_series_flag: bool = True

        if metric is None:
            metric: str = self.metrics[0]

        _cids: List[str] = self.cids if cids is None else cids
        _xcats: List[str] = self.xcats if xcats is None else xcats

        tickers_to_plot: List[str] = QuantamentalDataFrame(self.df).list_tickers()

        _tk: List[str] = tickers_to_plot.copy()
        if compare_series:
            _tk.remove(compare_series)
        grid_dim: Tuple[int, int] = self._get_grid_dim(
            tickers=_tk,
            ncols=ncols,
            attempt_square=attempt_square,
            cid_grid=cid_grid,
            xcat_grid=xcat_grid,
            cid_xcat_grid=cid_xcat_grid,
            _cids=cids,
            _xcats=xcats,
        )

        # if the facet size is not none, re-calc the figsize
        if facet_size is not None:
            figsize: Tuple[float, float] = (
                grid_dim[0] * facet_size[0] * 1.5,
                grid_dim[1] * facet_size[1] * 1.5,
            )
            # mul by 1.5 to account for the space taken up annotations, etc.
            # fig.tight_layout() cleans this up

        # form a dictionary of what goes on each plot
        # each key is an int - identifying the plot (L-R, T-B)
        # each value is a dict with keys "xs", "ys"

        if compare_series is not None:
            # check that the compare series exists in the dataframe
            if compare_series not in tickers_to_plot:
                raise ValueError(
                    f"Compare series {compare_series} not found in dataframe."
                )
            else:
                tickers_to_plot.remove(compare_series)

        plot_dict: Dict[str, Dict[str, Union[str, List[str]]]] = {}

        colormap = plt.cm.tab10
        legend_color_map: Optional[Dict[str, str]] = None

        if facet_titles is None:
            if cid_grid:
                facet_titles: List[str] = _cids
            elif xcat_grid:
                facet_titles: List[str] = _xcats
            elif cid_xcat_grid:
                # cid_xcat_grid facets only make sense if they have cid_xcat as the title
                legend: bool = False
            else:
                facet_titles: List[str] = tickers_to_plot
        elif isinstance(facet_titles, dict):
            if cid_grid:
                if not all([x in facet_titles.keys() for x in _cids]):
                    raise ValueError(
                        "Facet titles dictionary does not contain all cids."
                    )
            elif xcat_grid:
                if not all([x in facet_titles.keys() for x in _xcats]):
                    raise ValueError(
                        "Facet titles dictionary does not contain all xcats."
                    )

        if not any([cid_grid, xcat_grid, cid_xcat_grid]):
            # each ticker gets its own plot
            for i, ticker in enumerate(tickers_to_plot):
                tks: List[str] = [ticker]
                if compare_series is not None:
                    tks.append(compare_series)
                plot_dict[i] = {
                    "X": "real_date",
                    "Y": tks,
                }
                # plot_dict[i] --> Dict[str, Union[str, List[str]]]

        if cid_grid or xcat_grid:
            # flipper handles resolution between cid_grid and xcat_grid for binary
            # variables
            flipper: bool = 1 if cid_grid else -1
            if facet_titles is None:
                # needs to be "flipped" twice, as facet_titles need to be complementary
                # to the legend labels
                facet_titles: List[str] = [_cids, _xcats][::flipper][0]
            if legend_labels is None:
                legend_labels: List[str] = [_xcats, _cids][::flipper][0]
            elif len(legend_labels) != (
                len([_xcats, _cids][::flipper][0]) + bool(compare_series)
            ):
                raise ValueError(
                    "The number of legend labels does not match the lines to plot."
                )

            if legend_color_map is None:
                legend_color_map: Dict[str, str] = {
                    x: colormap(i) for i, x in enumerate([_xcats, _cids][::flipper][0])
                }

            for i, fvar in enumerate([_cids, _xcats][::flipper][0]):
                tks: List[str] = [
                    "_".join([fvar, x][::flipper])
                    for x in ([_xcats, _cids][::flipper][0])
                ]

                if tks == [compare_series]:
                    continue

                if isinstance(facet_titles, dict):
                    facet_title = facet_titles[fvar]
                else:
                    facet_title = facet_titles[i]

                plot_dict[i] = {
                    "X": "real_date",
                    "Y": tks + ([compare_series] if compare_series else []),
                    "title": facet_title,  # facet_titles
                }
                # plot_dict[i] --> Dict[str, Union[str, List[str]]]

        if cid_xcat_grid:
            # NB : legend goes away in cid_xcat_grid
            legend: bool = False

            for i, cid in enumerate(_cids):
                for j, xcat in enumerate(_xcats):
                    tk: str = "_".join([cid, xcat])
                    plot_dict[i * len(_xcats) + j] = {
                        "X": "real_date",
                        "Y": [tk],
                        "title": tk,
                    }
                    # plot_dict[i * len(_xcats) + j] --> Dict[str, List[str]]

        if not any([cid_grid, xcat_grid, cid_xcat_grid]):
            for i, (key, plt_dct) in enumerate(plot_dict.items()):
                plt_dct["title"] = plt_dct["Y"][0] + " vs " + plt_dct["X"][0]

        if len(plot_dict) == 0:
            raise ValueError("Unable to resolve plot settings.")

        # sort by the title - only
        if all("title" in ditem.keys() for ditem in plot_dict.values()):
            _plot_dict: Dict[str, Dict[str, Union[str, List[str]]]] = {
                i: ditem for i, ditem in enumerate(plot_dict.values())
            }
            plot_dict: Dict[str, Dict[str, Union[str, List[str]]]] = _plot_dict.copy()

        ##############################
        # Plotting
        ##############################

        fig = plt.figure(figsize=figsize)
        renderer = fig.canvas.get_renderer()

        outer_gs: GridSpec = GridSpec(
            ncols=grid_dim[0],
            nrows=grid_dim[1],
            figure=fig,
        )

        # re_adj: List[float] = [Left, Bottom, Right, Top]
        re_adj: List[float] = [0, 0, 1, 1]

        suptitle = fig.suptitle(
            title,
            fontsize=title_fontsize,
            x=title_xadjust,
            y=title_yadjust,
        )

        if title is not None:
            # get the figure coordinates of the title
            fig_width, fig_height = (
                suptitle.get_window_extent(renderer=renderer).width,
                suptitle.get_window_extent(renderer=renderer).height,
            )
            title_newline_adjust: float = title_fontsize / 500
            if title_yadjust is not None and title_yadjust != 1.0:
                title_newline_adjust = abs(title_yadjust - 1.0)
            # count the number of newlines in the title

            title_height: float = (title.count("\n") + 1) * title_newline_adjust

            re_adj[3] = (
                re_adj[3]
                - title_height
                + fig_height / fig.get_window_extent(renderer=renderer).height
            )

        axs: Union[np.ndarray, plt.Axes] = outer_gs.subplots(
            sharex=share_x,
            sharey=False,
        )

        if not isinstance(axs, np.ndarray):
            axs = np.array([axs])
        ax_list = axs.flatten().tolist()

        self.df.set_index(["cid", "xcat"], inplace=True)
        self.df.sort_index(inplace=True)

        for i, (key, plt_dct) in enumerate(plot_dict.items()):
            ax_i = ax_list[i]

            if plt_dct["X"] != "real_date":
                raise NotImplementedError(
                    "Only `real_date` is supported for the X axis."
                )

            is_empty_plot = False

            if plot_func is not None:
                plot_func(df=self.df, plt_dict=plt_dct, ax=ax_i, **plot_func_kwargs)
                continue

            for iy, y in enumerate(plt_dct["Y"]):
                cidx, xcatx = str(y).split("_", 1)
                try:
                    selected_df = self.df.loc[cidx, xcatx]
                    is_valid_series = True
                except KeyError:
                    is_valid_series = False
                is_empty_plot = is_empty_plot and not is_valid_series

                if is_valid_series:
                    X = selected_df[plt_dct["X"]].values
                    Y = selected_df[metric].values
                else:
                    X, Y = [], []

                plot_func_args = {}
                if legend_color_map:
                    plot_func_args["color"] = legend_color_map.get(
                        xcatx if cid_grid else cidx
                    )

                if y == compare_series:
                    plot_func_args["color"] = "red"
                    plot_func_args["linestyle"] = "--"

                if not interpolate:
                    X, Y = self._insert_nans(X, Y)

                if plot_func is None:
                    ax_i.plot(X, Y, **plot_func_args)
                else:
                    plot_func(real_dates=X, values=Y, ax=ax_i, **plot_func_kwargs)

            if not cid_xcat_grid:
                if facet_titles:
                    ax_i.set_title(
                        plt_dct["title"],
                        fontsize=facet_title_fontsize,
                        x=facet_title_xadjust,
                        y=facet_title_yadjust,
                    )
                if x_axis_label is not None:
                    ax_i.set_xlabel(x_axis_label, fontsize=axis_fontsize)
                if y_axis_label is not None:
                    ax_i.set_ylabel(y_axis_label, fontsize=axis_fontsize)
            else:
                if i < grid_dim[0]:
                    ax_i.set_title(
                        plt_dct["Y"][0].split("_", 1)[1],
                        fontsize=axis_fontsize,
                    )
                if i % grid_dim[0] == 0:
                    ax_i.set_ylabel(
                        plt_dct["Y"][0].split("_", 1)[0],
                        fontsize=axis_fontsize,
                    )

            if is_empty_plot:
                ax_i.set_xticks([])
                ax_i.set_yticks([])
            else:
                if ax_grid:
                    ax_i.grid(axis="both", linestyle="--", alpha=0.5)
                if ax_hline is not None:
                    ax_i.axhline(ax_hline, color="black", linestyle="--")
                if ax_vline is not None:
                    raise NotImplementedError(
                        "Vertical axis lines are not supported at this time."
                    )

        self.df.reset_index(inplace=True)

        for ax in ax_list[len(plot_dict) :]:
            fig.delaxes(ax)
        ax_list = ax_list[: len(plot_dict)]

        if share_x:
            for target_ax in ax_list[(len(plot_dict) - grid_dim[0]) :]:
                target_ax.xaxis.set_tick_params(labelbottom=True)

        if share_y:
            ymins, ymaxs = zip(*(ax.get_ylim() for ax in ax_list))
            global_min, global_max = min(ymins), max(ymaxs)
            abs_max = max(abs(global_min), abs(global_max))
            global_min, global_max = -abs_max, abs_max
            for ax in ax_list:
                ax.set_ylim(global_min, global_max)

            width, height = grid_dim
            for idx, ax in enumerate(ax_list):
                col_idx = idx % width
                if col_idx != 0:
                    ax.tick_params(labelleft=False)

        if legend:
            if "lower" in legend_loc and legend_ncol < grid_dim[0]:
                legend_ncol = grid_dim[0]

            leg = fig.legend(
                labels=legend_labels,
                loc=legend_loc,
                fontsize=legend_fontsize,
                ncol=legend_ncol,
                bbox_to_anchor=legend_bbox_to_anchor,
                frameon=legend_frame,
            )

            leg_width, leg_height = (
                leg.get_window_extent(renderer=renderer).width,
                leg.get_window_extent(renderer=renderer).height,
            )
            fig_width, fig_height = (
                fig.get_window_extent(renderer=renderer).width,
                fig.get_window_extent(renderer=renderer).height,
            )

            if "lower" in legend_loc:
                re_adj[1] += leg_height / fig_height
            if "upper" in legend_loc:
                re_adj[3] -= leg_height / fig_height
            if "left" in legend_loc:
                re_adj[0] += leg_width / fig_width
            if "right" in legend_loc:
                re_adj[2] -= leg_width / fig_width

        outer_gs.tight_layout(fig, rect=re_adj)

        if save_to_file is not None:
            fig.savefig(save_to_file, dpi=dpi)

        if show:
            plt.show()

        if return_figure:
            return fig

    def _insert_nans(self, X: Union[np.ndarray, List], Y: Union[np.ndarray, List]):
        """
        Inserts a single NaN value in each inferred gap in the time series.

        Parameters
        ----------
        X : ~numpy.ndarray
            Array of dates.
        Y : ~numpy.ndarray
            Array of values.

        Returns
        -------
        Tuple[~numpy.ndarray, ~numpy.ndarray]
            Tuple of arrays with gaps filled.
        """

        if len(X) == 0 or len(Y) == 0:
            return X, Y

        df = pd.DataFrame({"X": X, "Y": Y})
        df["X"] = pd.to_datetime(df["X"])

        # Infer the most common frequency of the time series
        differences = df["X"].diff().dt.total_seconds().dropna().astype(int)
        frequency_seconds = differences.mode()[0]

        frequency = pd.to_timedelta(frequency_seconds, unit="s")

        # Ignore weekends if daily
        freq_threshold = max(frequency_seconds * 1.5, 60 * 60 * 24 * 3)
        gaps = differences > freq_threshold
        gap_positions = gaps[gaps].index.tolist()

        new_rows = []
        for pos in gap_positions:
            gap_date = df.loc[pos - 1, "X"] + frequency
            new_rows.append({"X": gap_date, "Y": np.nan})

        if new_rows:
            new_rows_df = pd.DataFrame(new_rows)
            df = (
                pd.concat([df, new_rows_df], ignore_index=True)
                .sort_values(by="X")
                .reset_index(drop=True)
            )

        return df.X, df.Y


if __name__ == "__main__":
    # from macrosynergy.visuals import FacetPlot
    import time

    from macrosynergy.management.simulate import make_test_df

    np.random.seed(0)

    cids_A: List[str] = ["AUD", "CAD", "EUR", "GBP", "USD"]
    cids_B: List[str] = ["CHF", "INR", "JPY", "NOK", "NZD", "SEK"]
    cids_C: List[str] = ["CHF", "EUR", "INR", "JPY", "NOK", "NZD", "SEK", "USD"]

    xcats_A: List[str] = [
        "CPIXFE_SA_P1M1ML12",
        "CPIXFE_SJA_P3M3ML3AR",
        "CPIXFE_SJA_P6M6ML6AR",
        "CPIXFE_SA_P1M1ML12_D1M1ML3",
        "CPIXFE_SA_P1M1ML12_D1M1ML3",
    ]
    xcats_B: List[str] = [
        "CPIC_SA_P1M1ML12",
        "CPIC_SJA_P3M3ML3AR",
        "CPIC_SJA_P6M6ML6AR",
        "CPIC_SA_P1M1ML12_D1M1ML3",
        "CPIH_SA_P1M1ML12",
        "EXALLOPENNESS_NSA_1YMA",
        "EXMOPENNESS_NSA_1YMA",
    ]
    xcats_C: List[str] = ["DU05YXR_NSA", "DU05YXR_VT10"]
    xcats_D: List[str] = [
        "FXXR_NSA",
        "EQXR_NSA",
        "FXTARGETED_NSA",
        "FXUNTRADABLE_NSA",
    ]
    all_cids: List[str] = sorted(list(set(cids_A + cids_B + cids_C)))
    all_xcats: List[str] = sorted(list(set(xcats_A + xcats_B + xcats_C + xcats_D)))

    df: pd.DataFrame = make_test_df(
        cids=all_cids,
        xcats=all_xcats,
    )

    # remove data for USD_FXXR_NSA and CHF _EQXR_NSA and _FXXR_NSA
    df: pd.DataFrame = df[
        ~((df["cid"] == "USD") & (df["xcat"] == "FXXR_NSA"))
    ].reset_index(drop=True)
    df: pd.DataFrame = df[
        ~((df["cid"] == "CHF") & (df["xcat"].isin(["EQXR_NSA", "FXXR_NSA"])))
    ].reset_index(drop=True)
    df: pd.DataFrame = df[
        ~((df["cid"] == "NOK") & (df["xcat"] == "FXUNTRADABLE_NSA"))
    ].reset_index(drop=True)

    timer_start: float = time.time()

    with FacetPlot(
        df,
    ) as fp:
        fp.lineplot(
            cids=cids_A,
            share_x=True,
            xcat_grid=True,
            ncols=2,
            title=(
                "Test Title with a very long title to see how it looks, \n and a "
                "new line - why not?"
            ),
            # save_to_file="test_0.png",
            ax_hline=75,
            show=True,
        )

        fp.lineplot(
            cids=cids_B,
            xcats=xcats_A,
            attempt_square=True,
            share_y=True,
            cid_grid=True,
            title="Another test title",
            # save_to_file="test_1.png",
            show=True,
            share_x=True,
        )
        fp.lineplot(
            cids=cids_C,
            xcats=xcats_D,
            cid_xcat_grid=True,
            title="Another test title",
            show=True,
        )

    print(f"Time taken: {time.time() - timer_start}")
